"""
a tkinter frame combining 2D line plots and images from a NeXus file
"""
import tkinter as tk
from tkinter import ttk

import hdfmap

from ..misc.styles import create_root
from ..misc.logging import create_logger
from .nexus_plot import NexusMultiAxisPlot
from .nexus_image import NexusDetectorImage

logger = create_logger(__file__)


class NexusPlotAndImage(NexusMultiAxisPlot, NexusDetectorImage):
    """
    tkinter widget containing 2D line plot and image plot

    widget = NexusPlotAndImage(root, 'path/to/file.nxs', config=config)

    """
    def __init__(self, root: tk.Misc, *hdf_filenames: str,
                 config: dict | None = None, horizontal_alignment: bool = False):
        grid_options = dict(padx=5, pady=5, sticky='nsew')
        root.rowconfigure(0, weight=1)
        root.columnconfigure(0, weight=1)

        # 2D Plot
        frm = ttk.LabelFrame(root, text='Plot')
        frm.grid(column=0, row=0, **grid_options)
        sec = ttk.Frame(frm)
        sec.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES)
        # super().__init__(sec, *hdf_filenames, config=config)
        NexusMultiAxisPlot.__init__(self, sec, config=config)

        # Image
        frm = ttk.LabelFrame(root, text='Image')
        if horizontal_alignment:
            frm.grid(column=1, row=0, **grid_options)
        else:
            frm.grid(column=0, row=1, **grid_options)
        self.image_frame = ttk.Frame(frm)  # image frame will be packed when required
        NexusDetectorImage.__init__(self, self.image_frame, config=config)

        # Index line (used by both components)
        self.index_line, = self.ax1.plot([], [], ls='--', c='k', scaley=False, label=None)

        if hdf_filenames:
            self.update_data_from_files(*hdf_filenames)

    def pack_image(self):
        self.image_frame.pack(fill=tk.BOTH, expand=tk.YES)

    def remove_image(self):
        self.image_frame.pack_forget()

    def update_index_line(self):
        """update image_widget update_image to add plot line"""
        xvals, yvals = self.line.get_data()
        index = self.view_index.get()
        ylim = self.ax1.get_ylim()
        xval = xvals[index]
        self.index_line.set_data([xval, xval], ylim)
        self.update_axes()

    def update_data_from_files(self, *filenames: str, hdf_map: hdfmap.NexusMap | None = None):
        hdf_map = hdf_map or hdfmap.create_nexus_map(filenames[0])
        NexusMultiAxisPlot.update_data_from_files(self, *filenames, hdf_map=hdf_map)
        if hdf_map.image_data:
            NexusDetectorImage.update_image_data_from_file(self, filenames[0], hdf_map=hdf_map)
            self.update_index_line()
            self.pack_image()
        else:
            self.index_line.set_data([], [])
            self.remove_image()

    def update_image(self, event=None):
        super().update_image(event)
        self.update_index_line()

    def add_config_rois(self):
        super().add_config_rois()
        # add rois to signal drop-down
        for item in self.roi_names:
            self.listbox.insert("", tk.END, text=item)

    def new_window(self):
        window = create_root(self.filename, self.parent)
        widget = NexusPlotAndImage(window, config=self.config, horizontal_alignment=True)
        widget.update_data_from_files(self.filename, hdf_map=self.map)
        return widget



"""
widget for running scripts
"""

import os
import tkinter as tk
from tkinter import ttk

import hdfmap

from mmg_toolbox.scripts import scripts
from mmg_toolbox.utils.env_functions import get_first_file, run_python_script
from ..misc.logging import create_logger
from ..misc.config import get_config
from ..misc.functions import select_folder
from ..widgets.scan_range_selector import ScanRangeSelector

logger = create_logger(__file__)


class ScriptRunner:
    """Frame with """

    def __init__(self, root: tk.Misc, config: dict | None = None):
        logger.info('Creating ScriptRunner')
        self.root = root
        self.config = config or get_config()

        exp_directory = self.config.get('default_directory')
        proc_directory = self.config.get('processing_directory')

        self.exp_folder = tk.StringVar(root, exp_directory)
        self.proc_folder = tk.StringVar(root, proc_directory)
        self.script_name = tk.StringVar(root, 'example')
        self.notebook_name = tk.StringVar(root, 'example')
        self.script_desc = tk.StringVar(root, 'blah')
        self.notebook_desc = tk.StringVar(root, 'basd')
        self.output_file = tk.StringVar(root, proc_directory + '/file.py')
        self.metadata_name = tk.StringVar(self.root, '')
        self.options = {}
        self.file_list = []

        sec = ttk.LabelFrame(self.root, text='Folders')
        sec.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES, padx=4, pady=4)

        frm = ttk.Frame(sec)
        frm.pack(side=tk.TOP, fill=tk.X, expand=tk.YES, padx=4)
        ttk.Label(frm, text='Data Dir:', width=15).pack(side=tk.LEFT, padx=4)
        ttk.Entry(frm, textvariable=self.exp_folder, width=60).pack(side=tk.LEFT)
        ttk.Button(frm, text='Browse', command=self.browse_datadir).pack(side=tk.LEFT)

        frm = ttk.Frame(sec)
        frm.pack(side=tk.TOP, fill=tk.X, expand=tk.YES, padx=4)
        ttk.Label(frm, text='Analysis Dir:', width=15).pack(side=tk.LEFT, padx=4)
        ttk.Entry(frm, textvariable=self.proc_folder, width=60).pack(side=tk.LEFT)
        ttk.Button(frm, text='Browse', command=self.browse_analysis).pack(side=tk.LEFT)

        # Metadata selection
        sec = ttk.LabelFrame(self.root, text='Metadata')
        sec.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES, padx=4, pady=4)

        frm = ttk.Frame(sec)
        frm.pack(side=tk.TOP, fill=tk.X, expand=tk.YES, padx=4)
        ttk.Label(frm, text='Metadata:', width=15).pack(side=tk.LEFT, padx=4)
        ttk.Entry(frm, textvariable=self.metadata_name, width=60).pack(side=tk.LEFT)
        ttk.Button(frm, text='Choose', command=self.browse_metadata).pack(side=tk.LEFT)

        # Range selection
        sec = ttk.LabelFrame(self.root, text='Scan Numbers')
        sec.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES, padx=4, pady=4)

        self.range = ScanRangeSelector(sec, exp_directory, self.config)
        self.range.exp_folder.set(exp_directory)

        # Script Selection
        sec = ttk.LabelFrame(self.root, text='Script')
        sec.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES, padx=4, pady=4)

        line = ttk.Frame(sec)
        line.pack(side=tk.TOP, fill=tk.X, expand=tk.YES)
        var = ttk.OptionMenu(line, self.script_name, self.script_name.get(), *scripts.SCRIPTS.keys(),
                             command=self.script_select)
        var.pack(side=tk.LEFT, padx=4)
        ttk.Label(line, textvariable=self.script_desc).pack(side=tk.LEFT)

        sec = ttk.LabelFrame(self.root, text='Notebook')
        sec.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES, padx=4, pady=4)
        line = ttk.Frame(sec)
        line.pack(side=tk.TOP, fill=tk.X, expand=tk.YES)
        var = ttk.OptionMenu(line, self.notebook_name, self.notebook_name.get(), *scripts.NOTEBOOKS.keys(),
                             command=self.notebook_select)
        var.pack(side=tk.LEFT, padx=4)
        ttk.Label(line, textvariable=self.notebook_desc).pack(side=tk.LEFT)

        line = ttk.Frame(self.root)
        line.pack(side=tk.TOP, fill=tk.X, expand=tk.YES)
        ttk.Label(line, text='file', width=6).pack(side=tk.LEFT, padx=2)
        ttk.Entry(line, textvariable=self.output_file, width=60).pack(side=tk.LEFT, padx=2)
        ttk.Button(line, text='RUN', command=self.run_template).pack(side=tk.LEFT)
        ttk.Button(line, text='View', command=self.view_script).pack(side=tk.LEFT)

    def browse_metadata(self):
        from ..apps.namespace_select import create_metadata_selector
        scan_file = next(iter(self.range.generate_scan_files().values()), get_first_file(self.exp_folder.get()))
        hdf_map = hdfmap.create_nexus_map(scan_file)
        paths = create_metadata_selector(hdf_map)
        if paths:
            self.metadata_name.set(', '.join(path for path in paths))

    def update_options(self):
        exp_folder = self.exp_folder.get()
        self.range.exp_folder.set(exp_folder)
        scans = self.range.generate_scan_files()
        scan_numbers = list(scans)
        scan_files = scans.values()
        new = {
            # {{template}}: replacement
            'beamline': self.config.get('beamline', ''),
            # 'description': '',
            'filepaths': '\n    '.join(scan_files),
            'experiment_dir': exp_folder,
            'scan_numbers': str(scan_numbers),
            # 'title': 'a nice plot',
            'x-axis': 'axes',
            'y-axis': 'signal',
            'value': self.metadata_name.get()
        }
        self.options.clear()
        self.options.update(new)

    def script_select(self, event=None):
        filename, desc = scripts.SCRIPTS[self.script_name.get()]
        self.script_desc.set(desc)
        proc = self.proc_folder.get()
        self.output_file.set(os.path.join(proc, filename))

    def notebook_select(self, event=None):
        filename, desc = scripts.NOTEBOOKS[self.notebook_name.get()]
        self.notebook_desc.set(desc)
        proc = self.proc_folder.get()
        self.output_file.set(os.path.join(proc, filename))

    def run_template(self, event=None):
        if self.range.text.get("1.0", tk.END).strip() == '':
            return
        self.update_options()
        output_file = self.output_file.get()
        if output_file.endswith('.ipynb'):
            print(f"Creating notebook file: {output_file}")
            notebook_template = self.notebook_name.get()
            scripts.create_notebook(output_file, notebook_template, **self.options)
            print("Running notebook...")
        elif output_file.endswith('.py'):
            print(f"Creating script file: {output_file}")
            script_template = self.script_name.get()
            scripts.create_script(output_file, script_template, **self.options)
            print(f"Running script...")
            run_python_script(output_file)
        else:
            raise Exception('File is not script or notebook')

    def run_script(self, script_file: str):
        pass

    def run_notebook(self, notebook_file: str, *scan_files: str, **options):
        pass

    def view_script(self):
        from ..apps.python_editor import create_python_editor
        create_python_editor(open(self.output_file.get()).read(), parent=self.root, config=self.config)

    def browse_datadir(self):
        folder = select_folder(self.root)
        if folder:
            self.exp_folder.set(folder)

    def browse_analysis(self):
        folder = select_folder(self.root)
        if folder:
            self.proc_folder.set(folder)




"""
Plot Manager for the Scan object
"""

import numpy as np
import matplotlib.pyplot as plt
from ..nexus.nexus_scan import NexusScan
from .matplotlib import (
    set_plot_defaults, plot_line, plot_image,
    FIG_SIZE, FIG_DPI, DEFAULT_CMAP
)

"----------------------------------------------------------------------------------------------------------------------"
"----------------------------------------------- ScanPlotManager ------------------------------------------------------"
"----------------------------------------------------------------------------------------------------------------------"


class ScanPlotManager:
    """
    ScanPlotManager
        scan.plot = ScanPlotManager(scan)
        scan.plot() # plot default axes
        scan.plot.plot(xaxis, yaxis)  # creates figure
        scan.plot.plotline(xaxis, yaxis)  # plots line on current figure
        scan.plot.image()  # create figure and display detector image

    :param scan: NexusScan object
    """
    set_plot_defaults = set_plot_defaults

    def __init__(self, scan: NexusScan):
        self.scan = scan
        self.show = plt.show

    def __call__(self, *args, **kwargs):
        return self.plot(*args, **kwargs)

    def plotline(self, xaxis: str = 'axes', yaxis: str = 'signal', *args, **kwargs):
        """
        Plot scanned datasets on matplotlib axes subplot
        :param xaxis: str name or address of array to plot on x axis
        :param yaxis: str name or address of array to plot on y axis
        :param args: given directly to plt.plot(..., *args, **kwars)
        :param axes: matplotlib.axes subplot, or None to use plt.gca()
        :param kwargs: given directly to plt.plot(..., *args, **kwars)
        :return: list lines object, output of plot
        """
        data = self.scan.get_plot_data(xaxis, yaxis)

        if 'label' not in kwargs:
            kwargs['label'] = self.scan.label()
        axes = kwargs.pop('axes') if 'axes' in kwargs else plt.subplot()
        lines = plot_line(axes, data['x'], data['y'], None, *args, **kwargs)
        return lines

    def plot(self, xaxis: str = 'axes', yaxis: str | list[str] = 'signal', *args,
             axes: plt.Axes | None = None, **kwargs) -> plt.Axes:
        """
        Create matplotlib figure with plot of the scan
        :param xaxis: str name or address of array to plot on x axis
        :param yaxis: str name or address of array to plot on y axis, also accepts list of names for multiple lines
        :param args: given directly to plt.plot(..., *args, **kwars)
        :param axes: matplotlib.axes subplot, or None to create a figure
        :param kwargs: given directly to plt.plot(..., *args, **kwars)
        :return: axes object
        """
        #TODO: handle 2D grid scans
        axes = plt.subplot() if axes is None else axes

        if isinstance(yaxis, str):
            yaxis = [yaxis]

        for _yaxis in yaxis:
            data = self.scan.get_plot_data(xaxis, _yaxis)
            plot_line(axes, data['x'], data['y'], None, *args, label=_yaxis, **kwargs)

        # Add labels
        xlab, ylab = self.scan.map.generate_ids(xaxis, yaxis[0], modify_missing=False)
        axes.set_xlabel(xlab)
        axes.set_title(self.scan.title())
        if len(yaxis) == 1:
            axes.set_ylabel(ylab)
        else:
            axes.legend()
        return axes

    def image(self, index: int | tuple | slice | None = None, xaxis: str = 'axes',
              axes: plt.Axes | None = None, clim: tuple[float, float] | None = None,
              cmap: str = DEFAULT_CMAP, colorbar: bool = False, **kwargs) -> plt.Axes:
        """
        Plot image in matplotlib figure (if available)
        :param index: int, detector image index, 0-length of scan, if None, use centre index
        :param xaxis: name or address of xaxis dataset
        :param axes: matplotlib axes to plot on (None to create figure)
        :param clim: [min, max] colormap cut-offs (None for auto)
        :param cmap: str colormap name (None for auto)
        :param colorbar: False/ True add colorbar to plot
        :param kwargs: additional arguments for plot_detector_image
        :return: axes object
        """
        # x axis data
        xdata = self.scan.eval(xaxis)
        xname, = self.scan.map.generate_ids(xaxis, modify_missing=False)
        xdata = np.reshape(xdata, -1)  # handle multi-dimension data

        # image data
        im = self.scan.image(index)
        if im is None:
            im = np.zeros((101, 101))
        if index is None or index == 'sum':
            xvalue = xdata[np.size(xdata) // 2]
        else:
            xvalue = xdata[index]

        # plot
        axes = plt.subplot() if axes is None else axes
        plot_image(axes, im, clim=clim, cmap=cmap, **kwargs)
        if not self.scan.map.image_data:
            axes.text(0.5, 0.5, 'No Detector Image', c='w',
                      horizontalalignment='center',
                      verticalalignment='center',
                      transform=axes.transAxes)
        if colorbar:
            plt.colorbar(ax=axes)
        ttl = '%s\n%s [%s] = %s' % (self.scan.title(), xname, index, xvalue)
        axes.set_title(ttl)
        return axes

    def detail(self, xaxis: str = 'axes', yaxis: str | list[str] = 'signal',
               index: int | tuple | slice | None = None, clim: tuple[float, float] | None = None,
               cmap: str = DEFAULT_CMAP, **kwargs) -> plt.Figure:
        """
        Create matplotlib figure with plot of the scan and detector image
        :param xaxis: str name or address of array to plot on x axis
        :param yaxis: str name or address of array to plot on y axis, also accepts list of names for multiple plots
        :param index: int, detector image index, 0-length of scan, if None, use centre index
        :param clim: [min, max] colormap cut-offs (None for auto)
        :param cmap: str colormap name (None for auto)
        :param kwargs: given directly to plt.plot(..., *args, **kwars)
        :return: figure object
        """

        # Create figure
        fig, ((lt, rt), (lb, rb)) = plt.subplots(2, 2, figsize=[FIG_SIZE[0] * 1.2, FIG_SIZE[1] * 1.2], dpi=FIG_DPI)
        fig.subplots_adjust(hspace=0.35, left=0.1, right=0.95)

        # Top left - line plot
        self.plot(xaxis, yaxis, axes=lt, **kwargs)

        # Top right - image plot
        try:
            self.image(index, xaxis, cmap=cmap, clim=clim, axes=rt)
        except (FileNotFoundError, KeyError, TypeError):
            rt.text(0.5, 0.5, 'No Image')
            rt.set_axis_off()

        # Bottom-Left - details
        details = str(self.scan)
        lb.text(-0.2, 1, details, ha='left', va='top', multialignment="left", fontsize=12, wrap=True)
        lb.set_axis_off()

        # Bottom-Right - fit results
        rb.set_axis_off()
        if 'fit' in yaxis:
            fit_report = str(self.scan.fit)
            rb.text(-0.2, 1, fit_report, ha='left', va='top',  multialignment="left", fontsize=12, wrap=True)
        return fig

    def scananddetector(self, xaxis: str = 'axes', yaxis: str | list[str] = 'signal',
                        index: int | tuple | slice | None = None, clim: tuple[float, float] | None = None,
                        cmap: str = DEFAULT_CMAP, **kwargs) -> plt.Figure:
        """
        Create matplotlib figure with plot of the scan and detector image
        :param xaxis: str name or address of array to plot on x axis
        :param yaxis: str name or address of array to plot on y axis, also accepts list of names for multiple plots
        :param index: int, detector image index, 0-length of scan, if None, use centre index
        :param clim: [min, max] colormap cut-offs (None for auto)
        :param cmap: str colormap name (None for auto)
        :param kwargs: given directly to plt.plot(..., *args, **kwars)
        :return: Figure object
        """

        # Create figure
        fig, (lt, rt) = plt.subplots(1, 2, figsize=[FIG_SIZE[0] * 1.5, FIG_SIZE[1]], dpi=FIG_DPI)
        fig.subplots_adjust(hspace=0.35, left=0.1, right=0.95)

        # left - line plot
        self.plot(xaxis, yaxis, axes=lt, **kwargs)

        # right - image plot
        self.image(index, xaxis, cmap=cmap, clim=clim, axes=rt)
        return fig

    def image_histogram(self, index: int | tuple | slice | None = None,
                        axes: plt.Axes | None = None, **kwargs) -> plt.Axes:
        """
        Plot image in matplotlib figure (if available)
        :param index: int, detector image index, 0-length of scan, if None, use centre index
        :param axes: matplotlib axes to plot on (None to create figure)
        :param kwargs: additional arguments for plot_detector_image
        :param cut_ratios: list of cut-ratios, each cut has a different colour and given as ratio of max intensity
        :return: axes object
        """
        if index is None:
            index = ()
        vol = self.scan.get_image(index=index)

        axes.hist(np.log10(vol[vol > 0].flatten()), 100)

        axes.set_xlabel('Log$_{10}$ Pixel Intensity')
        axes.set_ylabel('N')
        axes.set_title(self.scan.title())
        return axes





"""
mmg_toolbox tests
Test experiment folder functions
"""


from mmg_toolbox.utils.experiment import Experiment
from mmg_toolbox.nexus.nexus_scan import NexusScan, NexusDataHolder
from . import only_dls_file_system
from .example_files import DIR

@only_dls_file_system
def test_file_loader():
    exp = Experiment(DIR + '/i16', DIR + '/i16/cm37262-1')
    scan = exp.scan(1109527)
    assert isinstance(scan, NexusDataHolder)
    scan_range = range(1032120, 1032130)
    scans = exp.scans(*scan_range)
    assert len([scn for scn in scans if isinstance(scn, NexusScan)]) == len(scans)

@only_dls_file_system
def test_check_scan():
    exp = Experiment(DIR + '/i16/cm37262-1', instrument='i16')
    out = exp.scan_str(0)
    print(out)
    assert 'Atten = 80' in out
    scan_range = range(1032120, 1032130)
    strings = exp.scans_str(*scan_range)
    print('\n'.join(strings))
    assert len(strings) == len(scan_range)


@only_dls_file_system
def test_plots():
    exp = Experiment(DIR + '/i16/cm37262-1', instrument='i16')
    exp.plot(1032120)
    # exp.plot_scans()







"""
Treeview widget for folders
"""

import os
import time
import tkinter as tk
from tkinter import ttk
from threading import Thread

from mmg_toolbox.utils.file_functions import list_files, list_path_time, display_timestamp, get_hdf_string, folder_summary
from mmg_toolbox.utils.env_functions import open_terminal
from ..misc.styles import create_hover
from ..misc.functions import select_folder, post_right_click_menu
from ..misc.jupyter import launch_jupyter_notebook
from ..misc.logging import create_logger
from .treeview import CanvasTreeview

logger = create_logger(__file__)


class FolderTreeViewFrame(CanvasTreeview):
    """Frame with TreeView and entry for Folders"""

    def __init__(self, extension: str, root: tk.Misc, initial_directory: str | None = None):
        self.root = root
        self.search_str = ""
        self.search_time = time.time()
        self.search_reset = 3.0  # seconds
        self._prev_folder = ''
        self.extensions = (extension, '.*')

        # Variables
        self.filepath = tk.StringVar(root, os.path.expanduser('~'))
        self.extension = tk.StringVar(root, extension)
        self.hdf_path = tk.StringVar(root, '')
        self.show_hidden = tk.BooleanVar(root, False)
        self.read_datasets = tk.BooleanVar(root, True)
        self.search_label = tk.StringVar(root, '')

        # Columns
        columns = [
            # (name, text, width, reverse, sort_col)
            ("#0", 'Folder', 200, False, None),
            ("modified", 'Modified', 150, True, "modified_time"),
            ('modified_time', 'Modified', 0, False, None),
            ("files", 'Files', 30, False, None),
            ("data", 'Data', 200, False, None),
            ("filepath", 'File Path', 0, False, None),
        ]

        # Build widgets
        self.ini_folderpath()
        super().__init__(root, *columns)
        self.tree.bind("<<TreeviewOpen>>", self.populate_files)
        self.tree.bind("<Double-1>", self.on_double_click)
        self.tree.bind("<Return>", self.on_double_click)
        self.tree.bind("<KP_Enter>", self.on_double_click)
        self.tree.bind('<KeyPress>', self.on_key_press)
        self.tree.bind("<Button-3>", self.right_click_menu())

        # search label
        frm = ttk.Frame(self.root)
        frm.pack(fill=tk.X, side=tk.TOP)
        ttk.Label(frm, textvariable=self.search_label, style='smallMsg.TLabel').pack()

        # Populate
        if initial_directory:
            self.filepath.set(initial_directory)
        self.populate_folders()

    "======================================================"
    "================= init functions ====================="
    "======================================================"

    def ini_folderpath(self):
        frm = ttk.Frame(self.root)
        frm.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

        var = ttk.Button(frm, text='Browse', command=self.browse_folder, width=8)
        var.pack(side=tk.LEFT)
        var = ttk.Button(frm, text=u'\u2302', command=self.home_folder, width=3)
        var.pack(side=tk.LEFT)
        var = ttk.Button(frm, text=u'\u2190', command=self.back_folder, width=3)
        var.pack(side=tk.LEFT)
        var = ttk.Button(frm, text=u'\u2191', command=self.up_folder, width=3)
        var.pack(side=tk.LEFT)
        var = ttk.Entry(frm, textvariable=self.filepath, width=30)
        var.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)
        var.bind('<Return>', self.populate_folders)
        var.bind('<KP_Enter>', self.populate_folders)

        # var = ttk.Checkbutton(frm, text='Show hidden', variable=self.show_hidden, command=self.populate_folders)
        # var.pack(side=tk.LEFT)

        var = ttk.Button(frm, text='Options', command=self.file_options)
        var.pack(side=tk.RIGHT)
        var = ttk.Button(frm, text='Search', command=self.search_options)
        var.pack(side=tk.RIGHT)

    "======================================================"
    "=============== populate functions ==================="
    "======================================================"

    def _get_extension(self):
        extension = self.extension.get()
        return '' if extension.lower() in ['.*', 'any'] else extension

    def set_folder(self, folder_path: str):
        self.filepath.set(folder_path)
        self.populate_folders()

    def populate_folders(self, event=None):
        path_time = list_path_time(self.filepath.get())
        self._delete_tree()
        # ('modified', 'modified_time', 'files', 'dataset', 'filepath')
        self.tree.insert("", tk.END, text="..", values=('', '', '', '', ''))
        hide_hidden = not self.show_hidden.get()
        for path, mtime in path_time:
            name_str = os.path.basename(path)
            if hide_hidden and name_str != '.' and (name_str.startswith('.') or name_str.startswith('_')):
                continue
            if path == '.':
                path = self.filepath.get()
            time_str = display_timestamp(mtime)
            self.tree.insert("", tk.END, text=name_str, values=(time_str, mtime, '', '', path))
        self.update_folder_nfiles()

    def update_folder_nfiles(self, event=None):
        """Update the number of files in each directory, as a seperate process"""

        extension = self._get_extension()

        def fun():
            for branch in self.tree.get_children():  # folders
                if not self.tree.winfo_exists():
                    return
                folder = self.tree.set(branch, 'filepath')
                nfiles = len(list_files(folder, extension=extension))
                self.tree.set(branch, 'files', nfiles)
                if nfiles > 0:  # add subdirectory for files
                    self.tree.insert(branch, tk.END)  # empty
                    self.tree.item(branch, open=False)
        th = Thread(target=fun)
        #TODO: add deamon?
        th.start()  # will run until complete, may error if TreeView is destroyed

    def populate_files(self, event=None):
        """Add list of files below folder on folder expand"""
        item = self.tree.focus()
        if not item:
            return
        nfiles = self.tree.set(item, 'files')  # number of hdf files in folder
        if not nfiles:
            return
        else:
            # if len(self.tree.get_children(item)) == 1:
            # remove old entries
            self.tree.delete(*self.tree.get_children(item))
            # add hdf files
            path = self.tree.set(item, 'filepath')
            files = list_files(path, self._get_extension())
            start_time = time.time()
            for file in files:
                timestamp = os.stat(file).st_mtime
                mtime = display_timestamp(timestamp)
                self.tree.insert(item, tk.END, text=os.path.basename(file), values=(mtime, timestamp, '', '', file))
            if self.read_datasets.get():
                self.update_datasets(self.hdf_path.get())
            logger.info(f"Expanding took {time.time() - start_time:.3g} s")

    def update_datasets(self, event=None):
        """Update dataset values column for hdf files under open folders"""
        pass

    "======================================================"
    "============= navigation functions ==================="
    "======================================================"

    def browse_folder(self):
        self._prev_folder = self.filepath.get()
        folder_directory = select_folder(parent=self.root)
        if folder_directory:
            self.set_folder(folder_directory)

    def home_folder(self):
        self._prev_folder = self.filepath.get()
        self.set_folder(os.path.expanduser("~"))

    def back_folder(self):
        if self._prev_folder:
            self.set_folder(self._prev_folder)

    def up_folder(self):
        self._prev_folder = self.filepath.get()
        self.set_folder(os.path.abspath(os.path.join(self.filepath.get(), '..')))

    def on_double_click(self, event=None):
        """Open a folder or open a file in a new window"""
        filename, foldername = self.get_filepath()
        if filename:
            pass
        else:
            self.filepath.set(foldername)
            self.populate_folders()

    "======================================================"
    "================= button functions ==================="
    "======================================================"

    def file_options(self):
        window, fun_close = create_hover(self.root)

        frm = ttk.Frame(window, borderwidth=2)
        frm.pack(side=tk.TOP, fill=tk.BOTH)
        ttk.Label(frm, text='Extension: ').pack(side=tk.LEFT)
        ttk.OptionMenu(frm, self.extension, self.extension.get(), *self.extensions).pack(side=tk.LEFT)

        frm = ttk.Frame(window, borderwidth=2)
        frm.pack(side=tk.TOP, fill=tk.BOTH)
        ttk.Checkbutton(
            frm, text='Show hidden files', variable=self.show_hidden, command=self.populate_folders
        ).pack(side=tk.LEFT)
        
        def close():
            self.update_datasets()
            fun_close()

        frm = ttk.Frame(window, borderwidth=10)
        frm.pack(side=tk.BOTTOM, fill=tk.BOTH)
        ttk.Button(frm, text='Close', command=close).pack(side=tk.LEFT, fill=tk.X, expand=tk.YES)
        
        return window

    def search_options(self):
        window, fun_close = create_hover(self.root)

        file_search = tk.StringVar(window, '')
        match_case = tk.BooleanVar(window, False)
        whole_word = tk.BooleanVar(window, False)

        frm = ttk.Frame(window, borderwidth=2)
        frm.pack(side=tk.TOP, fill=tk.BOTH)
        ttk.Label(frm, text='Filename: ').pack(side=tk.LEFT)
        ttk.Entry(frm, textvariable=file_search, width=20).pack(side=tk.LEFT)
        ttk.Checkbutton(frm, text='Match case: ', variable=match_case).pack(side=tk.LEFT)
        ttk.Checkbutton(frm, text='Whole word: ', variable=whole_word).pack(side=tk.LEFT)

        def search():
            self._search(
                query=file_search.get(),
                match_case=match_case.get(),
                whole_word=whole_word.get(),
            )

        frm = ttk.Frame(window, borderwidth=10)
        frm.pack(side=tk.BOTTOM, fill=tk.BOTH)
        ttk.Button(frm, text='Search', command=search).pack(side=tk.LEFT, fill=tk.X, expand=tk.YES, padx=5)
        ttk.Button(frm, text='Close', command=fun_close).pack(side=tk.LEFT, fill=tk.X, expand=tk.YES)

        return window

    "======================================================"
    "================ general functions ==================="
    "======================================================"

    def get_filepath(self) -> tuple[str, str]:
        """
        Return filepath and folderpath of current selection
        :returns filename: str full filepath or None if selection isn't a file
        :returns foldername: str folder path
        """
        filename = None
        foldername = self.filepath.get()
        for iid in self.tree.selection():
            item = self.tree.item(iid)
            if self.tree.set(iid, 'files') == '' and item['text'] != '..':  # item is a file
                filename = self.tree.set(iid, 'filepath')
                foldername = os.path.dirname(filename)
            else:  # item is a folder
                foldername = self.tree.set(iid, 'filepath')
        logger.debug(f"Selected item: filename='{filename}', foldername='{foldername}'")
        return filename, foldername

    def copy_path(self):
        filepath, folderpath = self.get_filepath()
        self.root.clipboard_clear()
        if filepath:
            self.root.clipboard_append(filepath)
        else:
            self.root.clipboard_append(folderpath)

    def open_terminal(self):
        filepath, folderpath = self.get_filepath()
        cmd = f"cd {folderpath}"
        open_terminal(cmd)

    def open_jupyter_notebook(self):
        filename, foldername = self.get_filepath()
        if filename:
            launch_jupyter_notebook('notebook', file=filename)
        else:
            launch_jupyter_notebook('notebook', directory=foldername)

    def _right_click_folder(self) -> tk.Menu:
        # right-click menu - folder options
        m_folder = tk.Menu(self.root, tearoff=0)
        m_folder.add_command(label="Copy path", command=self.copy_path)
        m_folder.add_command(label="Open Terminal", command=self.open_terminal)
        # TODO: add these
        # m_folder.add_command(label='Launch Jupyter', command=self.launch_jupyter)
        # m_folder.add_command(mode="Open Folder Datasets", command=self.menu_folder_files)
        # m_folder.add_command(mode="Open Folder Plots", command=self.menu_folder_plot)
        # # m_folder.add_command(mode="Display Contents", command=self.menu_folder_plot)
        m_folder.add_command(label="Display Summary", command=self.open_folder_summary)
        return m_folder

    def _right_click_file(self) -> tk.Menu:
        # right-click menu - file options
        m_file = tk.Menu(self.root, tearoff=0)
        #TODO: add these
        m_file.add_command(label="Copy path", command=self.copy_path)
        # m_file.add_command(mode="open Treeview", command=self.open_nexus_treeview)
        # m_file.add_command(mode="open Plot", command=self.open_nexus_plot)
        # m_file.add_command(mode="open Image", command=self.open_nexus_image)
        # m_file.add_command(mode="open Namespace", command=self.menu_namespace_gui)
        # m_file.add_command(mode="open Nexus Classes", command=self.menu_class_gui)
        return m_file

    def right_click_menu(self):
        logger.info('Creating right click menu')
        def menu_popup(event):
            # select item
            iid = self.tree.identify_row(event.y)
            if iid:
                self.tree.selection_set(iid)
                filename, foldername = self.get_filepath()
                if filename:
                    logger.debug(f"Right click menu created for file: {filename}")
                    menu = self._right_click_file()
                else:
                    logger.debug(f"Right click menu created for folder: {foldername}")
                    menu = self._right_click_folder()
                post_right_click_menu(menu, event.x_root, event.y_root)
        return menu_popup

    def _delete_tree(self):
        self.tree.delete(*self.tree.get_children())

    def _on_close(self):
        # self.root.unbind_all('<KeyPress>')
        self.root.destroy()

    def on_key_press(self, event):
        """any key press performs search of folders, selects first matching folder"""
        # return if clicked on entry box
        # event.widget == self.tree
        if str(event.widget).endswith('entry'):
            return
        # reset search str after reset time
        ctime = time.time()
        if ctime > self.search_time + self.search_reset:
            self.search_label.set('')
            self.search_str = ""
        # update search time, add key to query
        self.search_time = ctime
        self.search_str += event.char
        self.search_label.set(self.search_str)

        self.tree.selection_remove(self.tree.selection())
        # search folder list
        for branch in self.tree.get_children():  # folders
            folder = self.tree.item(branch)['text']
            if self.search_str in folder[:len(self.search_str)].lower():
            # if folder.lower().startswith(self.search_str):
                self.tree.selection_add(branch)
                self.tree.see(branch)
                break

    "======================================================"
    "=============== widget functions ====================="
    "======================================================"

    def open_nexus_treeview(self):
        filename, foldername = self.get_filepath()
        logger.info(f"Opening nexus viewer for filename: {filename}")
        if filename:
            from .. import create_nexus_viewer
            create_nexus_viewer(filename, parent=self.root)

    def open_nexus_plot(self):
        filename, foldername = self.get_filepath()
        logger.info(f"Opening nexus plot viewer for filename: {filename}")
        if filename:
            from ..apps.nexus import create_nexus_plotter
            create_nexus_plotter(filename, parent=self.root)

    def open_nexus_image(self):
        filename, foldername = self.get_filepath()
        logger.info(f"Opening nexus image viewer for filename: {filename}")
        if filename:
            from ..apps.nexus import create_nexus_image_plotter
            create_nexus_image_plotter(filename, parent=self.root)

    def open_folder_summary(self):
        filename, foldername = self.get_filepath()
        logger.info(f"Opening folder summary for filename: {filename}")
        if foldername:
            from mmg_toolbox.tkguis.apps.edit_text import EditText
            summary = folder_summary(foldername)
            EditText(summary, self.root).show()

    "======================================================"
    "================= misc functions ====================="
    "======================================================"

    def _search(self, query: str, match_case: bool = False, whole_word: bool = False,
                file_only: bool = False, data_only: bool = False):
        for branch in self.tree.get_children():  # folders
            # folder = self.tree.item(branch)['text']
            for leaf in self.tree.get_children(branch):  # files
                item = self.tree.item(leaf)
                if len(item['values']) < 3:
                    continue
                file = item['text']
                value = self.tree.set(leaf, 'data')
                if file_only:
                    test = file
                elif data_only:
                    test = value
                else:
                    test = f"{file} {value}"
                test = test if match_case else test.lower()
                test = test.split() if whole_word else test
                if query in test:
                    self.tree.selection_add(leaf)
                    self.tree.see(leaf)


class NexusFolderTreeViewFrame(FolderTreeViewFrame):
    def __init__(self, root: tk.Misc, initial_directory: str | None = None, hdf_path: str | None = None):
        logger.info('Creating NexusFolderTreeViewFrame')
        super().__init__('.nxs', root, initial_directory)
        self.extensions = ('.nxs', '.hdf', '.hdf5', '.h5', '.*')
        if hdf_path:
            self.hdf_path.set(hdf_path)

    def _right_click_file(self) -> tk.Menu:
        # right-click menu - file options
        m_file = super()._right_click_file()
        m_file.add_command(label="open Treeview", command=self.open_nexus_treeview)
        m_file.add_command(label="open Plot", command=self.open_nexus_plot)
        m_file.add_command(label="open Image", command=self.open_nexus_image)
        # m_file.add_command(mode="open Namespace", command=self.menu_namespace_gui)
        # m_file.add_command(mode="open Nexus Classes", command=self.menu_class_gui)
        return m_file

    def on_double_click(self, event=None):
        """Open a folder or open a file in a new window"""
        if not self.tree.focus():
            return
        iid = self.tree.focus()
        item = self.tree.item(iid)
        if self.tree.set(iid, 'files') == '' and item['text'] != '..':
            self.open_nexus_treeview()
        else:
            # item is a folder, open folder
            self._prev_folder = self.filepath.get()
            self.set_folder(os.path.abspath(os.path.join(self._prev_folder, item['text'])))

    def update_datasets(self, event=None):
        """Update dataset values column for hdf files under open folders"""

        def fn():
            hdf_path = self.hdf_path.get()
            self.tree.heading("data", text=hdf_path)
            for branch in self.tree.get_children():  # folders
                folder = self.tree.set(branch, 'filepath')
                for leaf in self.tree.get_children(branch):  # files
                    if not self.tree.winfo_exists():
                        return
                    file = self.tree.item(leaf)['text']
                    if file:
                        filename = str(os.path.join(folder, file))
                        value = get_hdf_string(filename, hdf_path, '...')
                        self.tree.set(leaf, 'data', value)

        th = Thread(target=fn)
        th.start()  # will run until complete, may error if TreeView is destroyed

    def file_options(self):
        from hdfmap import create_nexus_map
        window = super().file_options()

        filename, foldername = self.get_filepath()
        hdf_filename = tk.StringVar(window, filename if filename else '')
        dataset_name = tk.StringVar(window, '')

        def get_hdf_path(event):
            _filename = hdf_filename.get()
            _name = dataset_name.get()
            if _filename and _name:
                nexus_map = create_nexus_map(filename)
                self.hdf_path.set(
                    nexus_map.get_path(_name)
                )

        sec = ttk.LabelFrame(window, text='NeXus Files')
        sec.pack(side=tk.TOP, fill=tk.BOTH)

        frm = ttk.Frame(sec, borderwidth=2)
        frm.pack(side=tk.TOP, fill=tk.BOTH)
        ttk.Checkbutton(
            frm, text='Read dataset', variable=self.read_datasets
        ).pack(side=tk.LEFT)

        frm = ttk.Frame(sec, borderwidth=2)
        frm.pack(side=tk.TOP, fill=tk.BOTH)
        ttk.Label(frm, text='File:', width=12).pack(side=tk.LEFT)
        var = ttk.Entry(frm, textvariable=hdf_filename)
        var.pack(side=tk.LEFT)
        var.bind('<Return>', get_hdf_path)
        var.bind('<KP_Enter>', get_hdf_path)

        frm = ttk.Frame(sec, borderwidth=2)
        frm.pack(side=tk.TOP, fill=tk.BOTH)
        ttk.Label(frm, text='Name:', width=12).pack(side=tk.LEFT)
        var = ttk.Entry(frm, textvariable=dataset_name)
        var.pack(side=tk.LEFT)
        var.bind('<Return>', get_hdf_path)
        var.bind('<KP_Enter>', get_hdf_path)

        frm = ttk.Frame(sec, borderwidth=2)
        frm.pack(side=tk.TOP, fill=tk.BOTH)
        # ttk.Button(frm, text='NeXus Dataset Path', command=self.select_dataset).pack(side=tk.LEFT)
        ttk.Label(frm, text='Path:', width=12).pack(side=tk.LEFT)
        var = ttk.Entry(frm, textvariable=self.hdf_path)
        var.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)
        var.bind('<Return>', self.update_datasets)
        var.bind('<KP_Enter>', self.update_datasets)

    def search_options(self):
        window, fun_close = create_hover(self.root)

        file_search = tk.StringVar(window, '')
        match_case = tk.BooleanVar(window, False)
        whole_word = tk.BooleanVar(window, False)

        dataset_search = tk.StringVar(window, '')
        ds_match_case = tk.BooleanVar(window, False)
        ds_whole_word = tk.BooleanVar(window, False)

        frm = ttk.Frame(window, borderwidth=2)
        frm.pack(side=tk.TOP, fill=tk.BOTH)
        ttk.Label(frm, text='Filename: ', width=12).pack(side=tk.LEFT)
        ttk.Entry(frm, textvariable=file_search, width=20).pack(side=tk.LEFT)
        ttk.Checkbutton(frm, text='Match case: ', variable=match_case).pack(side=tk.LEFT)
        ttk.Checkbutton(frm, text='Whole word: ', variable=whole_word).pack(side=tk.LEFT)

        sec = ttk.LabelFrame(window, text='NeXus', borderwidth=4)
        sec.pack(side=tk.TOP, fill=tk.BOTH)

        frm = ttk.Frame(sec, borderwidth=2)
        frm.pack(side=tk.TOP, fill=tk.BOTH)
        ttk.Label(frm, text='Dataset: ', textvariable=self.hdf_path).pack(side=tk.LEFT)

        frm = ttk.Frame(sec, borderwidth=2)
        frm.pack(side=tk.TOP, fill=tk.BOTH)
        ttk.Label(frm, text='Search: ', width=12).pack(side=tk.LEFT)
        ttk.Entry(frm, textvariable=dataset_search, width=20).pack(side=tk.LEFT)
        ttk.Checkbutton(frm, text='Match case: ', variable=ds_match_case).pack(side=tk.LEFT)
        ttk.Checkbutton(frm, text='Whole word: ', variable=ds_whole_word).pack(side=tk.LEFT)

        def search():
            self._search(
                query=file_search.get(),
                match_case=match_case.get(),
                whole_word=whole_word.get(),
                file_only=True,
            )

        def ds_search():
            self._search(
                query=dataset_search.get(),
                match_case=ds_match_case.get(),
                whole_word=ds_whole_word.get(),
                data_only=True,
            )

        frm = ttk.Frame(window, borderwidth=10)
        frm.pack(side=tk.BOTTOM, fill=tk.BOTH)
        ttk.Button(frm, text='Dataset\nSearch', command=ds_search).pack(side=tk.LEFT, fill=tk.X, expand=tk.YES, padx=5)
        ttk.Button(frm, text='Search', command=search).pack(side=tk.LEFT, fill=tk.X, expand=tk.YES, padx=5)
        ttk.Button(frm, text='Close', command=fun_close).pack(side=tk.LEFT, fill=tk.X, expand=tk.YES)

        return window


class JupyterFolderTreeViewFrame(FolderTreeViewFrame):
    def __init__(self, root: tk.Misc, initial_directory: str | None = None):
        logger.info('Creating JupyterFolderTreeViewFrame')
        super().__init__('.ipynb', root, initial_directory)
        self.tree.configure(displaycolumns=('modified', 'files'))  # hide columns

    def _right_click_file(self) -> tk.Menu:
        # right-click menu - file options
        m_file = super()._right_click_file()
        m_file.add_command(label="start jupyter", command=self.open_jupyter_notebook)
        m_file.add_command(label="view html", command=self.open_jupyter_html)
        return m_file


    def open_jupyter_html(self):
        from mmg_toolbox.utils.nb_runner import view_jupyter_notebook, view_notebook_html

        filename, foldername = self.get_filepath()
        if filename:
            html = filename.replace('.ipynb', '.html')
            if os.path.isfile(html):
                view_jupyter_notebook(html)
            else:
                # generate html in TMP
                view_notebook_html(filename)





"""
a tkinter frame with an image with slider
"""

import tkinter as tk
from tkinter import ttk

from ..misc.styles import create_root
from ..misc.matplotlib import ini_image, DEFAULT_COLORMAP, COLORMAPS
from ..misc.logging import create_logger

logger = create_logger(__file__)


AXES = ['axis 1', 'axis 2', 'axis 3']


class SimpleImage:
    """
    Simple Image plot
    """

    def __init__(self, root: tk.Misc, xdata, ydata, xlabel='', ylabel='', title=''):
        self.root = root
        self.fig, self.ax1, self.plot_list, self.ax1_image, self.cb1, self.toolbar = ini_image(self.root)
        self.ax1.set_xlabel(xlabel)
        self.ax1.set_ylabel(ylabel)
        self.ax1.set_title(title)
        self.plot(xdata, ydata)

    def plot(self, *args, **kwargs):
        lines = self.ax1.imshow(*args, **kwargs)
        self.plot_list.extend(lines)
        self.update_plot()

    def remove_lines(self):
        for obj in self.plot_list:
            obj.remove()
        self.plot_list.clear()

    def reset_plot(self):
        # self.ax1.set_xlabel(self.xaxis.get())
        # self.ax1.set_ylabel(self.yaxis.get())
        self.ax1.set_title('')
        self.ax1.set_prop_cycle(None)  # reset colours
        self.ax1.legend([]).set_visible(False)
        for obj in self.ax1.lines:
            obj.remove()

    def update_plot(self):
        self.ax1.relim()
        self.ax1.autoscale(True)
        self.ax1.autoscale_view()
        self.fig.canvas.draw()
        self.toolbar.update()


class ImageViewer:
    """
    HDF Image Viewer - display a 3+D dataset as a series of images
    Usage:
        HDFViewer("hdf_file.hdf")
    Select a dataset address (one will be choosen by default)
    Use the displayed slider and options to view the data

    :param hdf_filename: str filename of HDF file
    :param parent: tk root or None
    """

    def __init__(self, hdf_filename="", parent=None):

        self.root = create_root('HDF Image Viewer', parent=parent)

        # Variables
        self._ax = 0
        self._x_axis = 1
        self._y_axis = 2
        self.view_axis = tk.StringVar(self.root, AXES[self._ax])
        self.view_index = tk.IntVar(self.root, 0)
        self.axis_name = tk.StringVar(self.root, 'axis = ')
        self.axis_value = tk.DoubleVar(self.root, 0)
        self.add_phase = tk.DoubleVar(self.root, 0)
        self.logplot = tk.BooleanVar(self.root, False)
        self.difplot = tk.BooleanVar(self.root, False)
        self.mask = tk.DoubleVar(self.root, 0)
        self.cmin = tk.DoubleVar(self.root, 0)
        self.cmax = tk.DoubleVar(self.root, 1)
        self.fixclim = tk.BooleanVar(self.root, False)
        self.colormap = tk.StringVar(self.root, DEFAULT_COLORMAP)

        "----------- Options -----------"
        self.ini_options()

        "----------- Slider -----------"
        self.tkscale = self.ini_slider()

        "----------- Image -----------"
        self.fig, self.ax1, self.plot_list, self.ax1_image, self.cb1, self.toolbar = ini_image(self.root)

        "-------- Start Mainloop ------"
        if parent is None:
            self.root.mainloop()

    "======================================================"
    "================= init functions ====================="
    "======================================================"

    def ini_options(self):
        frm = ttk.LabelFrame(self.root, text='Options', relief=tk.RIDGE)
        frm.pack(expand=tk.NO, pady=2, padx=5)

        var = ttk.Checkbutton(frm, text='Log', variable=self.logplot, command=self.update_image)
        var.pack(side=tk.LEFT, padx=6)
        var = ttk.Checkbutton(frm, text='Diff', variable=self.difplot, command=self.update_image)
        var.pack(side=tk.LEFT, padx=6)

        var = ttk.Label(frm, text='Mask <')
        var.pack(side=tk.LEFT, expand=tk.NO, padx=6)
        var = ttk.Entry(frm, textvariable=self.mask, width=6)
        var.pack(side=tk.LEFT, padx=6)
        var.bind('<Return>', self.update_image)
        var.bind('<KP_Enter>', self.update_image)

        var = ttk.OptionMenu(frm, self.colormap, *COLORMAPS, command=self.update_image)
        var.pack(side=tk.LEFT)

        var = ttk.Label(frm, text='clim:')
        var.pack(side=tk.LEFT, expand=tk.NO)
        var = ttk.Entry(frm, textvariable=self.cmin, width=6)
        var.pack(side=tk.LEFT)
        var.bind('<Return>', self.update_image)
        var.bind('<KP_Enter>', self.update_image)
        var = ttk.Entry(frm, textvariable=self.cmax, width=6)
        var.pack(side=tk.LEFT)
        var.bind('<Return>', self.update_image)
        var.bind('<KP_Enter>', self.update_image)
        var = ttk.Checkbutton(frm, text='Fix', variable=self.fixclim)
        var.pack(side=tk.LEFT)

    def ini_slider(self):
        frm = ttk.Frame(self.root)
        frm.pack(expand=tk.NO, pady=2, padx=5)

        var = ttk.OptionMenu(frm, self.view_axis, None, *AXES, command=self.update_axis)
        var.pack(side=tk.LEFT)

        def inc():
            self.view_index.set(self.view_index.get() + 1)
            self.update_image()

        def dec():
            self.view_index.set(self.view_index.get() - 1)
            self.update_image()

        var = ttk.Label(frm, text='Index:', width=8)
        var.pack(side=tk.LEFT)
        var = ttk.Button(frm, text='-', command=dec)
        var.pack(side=tk.LEFT)
        tkscale = ttk.Scale(frm, from_=0, to=100, variable=self.view_index, orient=tk.HORIZONTAL,
                            command=self.update_image, length=300)
        # var.bind("<ButtonRelease-1>", callback)
        tkscale.pack(side=tk.LEFT, expand=tk.YES)
        var = ttk.Button(frm, text='+', command=inc)
        var.pack(side=tk.LEFT)
        var = ttk.Entry(frm, textvariable=self.view_index, width=6)
        var.pack(side=tk.LEFT)
        var.bind('<Return>', self.update_image)
        var.bind('<KP_Enter>', self.update_image)

        # axis mode
        var = ttk.Label(frm, textvariable=self.axis_name)
        var.pack(side=tk.LEFT)
        var = ttk.Label(frm, textvariable=self.axis_value)
        var.pack(side=tk.LEFT)
        return tkscale

    "======================================================"
    "================ general functions ==================="
    "======================================================"

    def _update_axis(self):
        """Get data size etc"""
        self._ax = int(self.view_axis.get()[-1]) - 1  # e.g. 'axis 1'

        hdf_filename = self.filepath.get()
        address = self.address.get()
        axis_address = self.axis_address.get()
        # Check dataset
        self.error_message = check_image_dataset(hdf_filename, address)
        if self.error_message:
            show_error(self.error_message, self.root)
        # Load image to get size and shape
        with load_hdf(hdf_filename) as hdf:
            dataset = hdf.get(address)

            shape = dataset_shape(dataset)
            if not self.fixclim.get():
                image = get_image(
                    dataset=dataset,
                    image_number=shape[self._ax]//2,
                    axis=self._ax
                )
                image_mean = np.mean(image[image > 0])
                image_max = np.max(image)
                cmax = image_mean + (image_max - image_mean) ** 0.7
                cmax = float(f"{cmax: .2g}")
                self.cmin.set(0)
                self.cmax.set(cmax)

            # axis values
            if axis_address:
                axis_dataset = hdf.get(axis_address)
                if axis_dataset and len(axis_dataset) == shape[self._ax]:
                    self.axis_name.set(f"{address_name(axis_address)} = ")
                    self.axis_value.set(axis_dataset[shape[self._ax]//2])

        # udpate scale and axes
        self.tkscale.config(to=shape[self._ax] - 1)  # set slider max
        if self._ax == 0:
            self.ax1.set_xlabel(u'Axis 3')
            self.ax1.set_ylabel(u'Axis 2')
            self.ax1.set_xlim([0, shape[2]])
            self.ax1.set_ylim([0, shape[1]])
        elif self._ax == 1:
            self.ax1.set_xlabel(u'Axis 3')
            self.ax1.set_ylabel(u'Axis 1')
            self.ax1.set_xlim([0, shape[2]])
            self.ax1.set_ylim([0, shape[0]])
        else:
            self.ax1.set_xlabel(u'Axis 2')
            self.ax1.set_ylabel(u'Axis 1')
            self.ax1.set_xlim([0, shape[1]])
            self.ax1.set_ylim([0, shape[0]])
        self.view_index.set(shape[self._ax]//2)

    def update_image(self, event=None):
        """Plot image data"""
        if self.error_message:
            show_error(self.error_message, self.root)
        # Load image
        image = get_hdf_image(
            hdf_filename=self.filepath.get(),
            address=self.address.get(),
            image_number=int(self.view_index.get()),
            axis=self._ax
        )
        # Options
        cmin, cmax = self.cmin.get(), self.cmax.get()
        if self.logplot.get():
            image = np.log10(image)
            cmax = np.log10(cmax)
        if self.difplot.get():
            raise Warning('Not implemented yet')
        if self.mask.get():
            raise Warning('Not implemented yet')
        # Add plot
        self.ax1_image.remove()
        colormap = self.colormap.get()
        clim = [cmin, cmax]
        self.ax1_image = self.ax1.pcolormesh(image, shading='auto', clim=clim, cmap=colormap)
        self.ax1_image.set_clim(clim)
        self.cb1.update_normal(self.ax1_image)
        self.toolbar.update()
        self.fig.canvas.draw()
        # Load axis mode
        value = get_hdf_array_value(
            hdf_filename=self.filepath.get(),
            address=self.axis_address.get(),
            image_number=int(self.view_index.get())
        )
        self.axis_value.set(value)

    "======================================================"
    "================= event functions ===================="
    "======================================================"

    def loadfile(self, event=None):
        self._loadfile(self.filepath.get())

    def update_axis(self, event=None):
        """Get data size etc"""
        self._update_axis()
        self.update_image()




"""
Fitting functions using lmfit

See: https://lmfit.github.io/lmfit-py/builtin_models.html

Use of peakfit:
from fitting import peakfit
fit = peakfit(xdata, ydata)  # returns lmfit object
print(fit)
fit.plot()
"""

import numpy as np
from lmfit.models import (
    GaussianModel, LorentzianModel, VoigtModel, PseudoVoigtModel, LinearModel, ExponentialModel,
    SineModel
)
from lmfit.model import ModelResult, Model, Parameters
from .misc_functions import stfm
from ..nexus.nexus_scan import NexusScan

__all__ = ['MODELS', 'PEAK_MODELS', 'BACKGROUND_MODELS', 'METHODS', 'poisson_errors', 'gauss', 'peak_ratio',
           'group_adjacent', 'find_peaks', 'find_peaks_str', 'peak_results', 'peak_results_str', 'peak_results_fit',
           'peak_results_plot', 'peakfit', 'modelfit', 'multipeakfit', 'peak2dfit', 'FitResults']

# https://lmfit.github.io/lmfit-py/builtin_models.html#peak-like-models
MODELS = {
    'gaussian': GaussianModel,
    'lorentz': LorentzianModel,
    'voight': VoigtModel,
    'pvoight': PseudoVoigtModel,
    'linear': LinearModel,
    'exponential': ExponentialModel,
    'SineModel': SineModel,
}  # list of available lmfit models

PEAK_MODELS = {
    'gaussian': ['gaussian', 'gauss'],
    'voight': ['voight', 'voight model'],
    'pvoight': ['pseudovoight', 'pvoight'],
    'lorentz': ['lorentz', 'lorentzian', 'lor'],
}  # alternaive names for peaks

BACKGROUND_MODELS = {
    'linear': ['flat', 'slope', 'linear', 'line', 'straight'],
    'exponential': ['exponential', 'curve']
}  # alternaive names for background models

# https://lmfit.github.io/lmfit-py/fitting.html#fit-methods-table
METHODS = {
    'leastsq': 'Levenberg-Marquardt',
    'nelder': 'Nelder-Mead',
    'lbfgsb': 'L-BFGS-B',
    'powell': 'Powell',
    'cg': 'Conjugate Gradient',
    'newton': 'Newton-CG',
    'cobyla': 'COBYLA',
    'bfgsb': 'BFGS',
    'tnc': 'Truncated Newton',
    'trust-ncg': 'Newton CG trust-region',
    'trust-exact': 'Exact trust-region',
    'trust-krylov': 'Newton GLTR trust-region',
    'trust-constr': 'Constrained trust-region',
    'dogleg': 'Dogleg',
    'slsqp': 'Sequential Linear Squares Programming',
    'differential_evolution': 'Differential Evolution',
    'brute': 'Brute force method',
    'basinhopping': 'Basinhopping',
    'ampgo': 'Adaptive Memory Programming for Global Optimization',
    'shgo': 'Simplicial Homology Global Ooptimization',
    'dual_annealing': 'Dual Annealing',
    'emcee': 'Maximum likelihood via Monte-Carlo Markov Chain',
}


def poisson_errors(y: np.ndarray) -> np.ndarray:
    """Default error function for counting statistics"""
    return np.sqrt(np.abs(y) + 1)


def peak_ratio(y: np.ndarray, yerror: np.ndarray | None = None) -> float:
    """
    Return the ratio signal / error for given dataset
    From Blessing, J. Appl. Cryst. (1997). 30, 421-426 Equ: (1) + (6)
      peak_ratio = (sum((y-bkg)/dy^2)/sum(1/dy^2)) / sqrt(i/sum(1/dy^2))
    :param y: array of y data
    :param yerror: array of errors on data, or None to calcualte np.sqrt(y+0.001)
    :return: float ratio signal / err
    """
    if yerror is None:
        yerror = poisson_errors(y)
    bkg = np.min(y)
    wi = 1 / yerror ** 2
    signal = np.sum(wi * (y - bkg)) / np.sum(wi)
    err = np.sqrt(len(y) / np.sum(wi))
    return signal / err


def gen_weights(yerrors=None) -> np.ndarray | None:
    """
    Generate weights for fitting routines
    :param yerrors: array(n) or None
    :return: array(n) or None
    """
    if yerrors is None or np.all(np.abs(yerrors) < 0.001):
        weights = None
    else:
        yerrors = np.asarray(yerrors, dtype=float)
        yerrors[yerrors < 1] = 1.0
        weights = 1 / yerrors
        weights = np.abs(np.nan_to_num(weights))
    return weights


def gauss(x: np.ndarray, y: np.ndarray | None = None, 
          height: float = 1, cen: float = 0, fwhm: float = 0.5, 
          bkg: float = 0, cen_y: float | None = None) -> np.ndarray:
    """
    Define Gaussian distribution in 1 or 2 dimensions

        y[1xn] = gauss(x[1xn], height=10, cen=0, fwhm=1, bkg=0)
           - OR -
        Z[nxm] = gauss(x[1xn], y[1xm], height=100, cen=4, fwhm=5, bkg=30)

    From http://fityk.nieto.pl/model.html


    :param x: [1xn] array of values, defines size of gaussian in dimension 1
    :param y: None* or [1xm] array of values, defines size of gaussian in dimension 2
    :param height: peak height
    :param cen: peak centre
    :param fwhm: peak full width at half-max
    :param bkg: background
    :param cen_y: peak centre in y-axis (None to use cen)
    :returns: [1xn array] Gassian distribution
    - or, if y is not None: -
    :returns: [nxm array] 2D Gaussian distribution
    """

    if cen_y is None:
        cen_y = cen
    if y is None:
        y = cen_y

    x = np.asarray(x, dtype=float).reshape([-1])
    y = np.asarray(y, dtype=float).reshape([-1])
    X, Y = np.meshgrid(x, y)
    g = height * np.exp(-np.log(2) * (((X - cen) ** 2 + (Y - cen) ** 2) / (fwhm / 2) ** 2)) + bkg

    if len(y) == 1:
        g = g.reshape([-1])
    return g


def group_adjacent(values: np.ndarray, close: float = 10):
    """
    Average adjacent values in array, return grouped array and indexes to return groups to original array
    
    E.G.
     grp, idx = group_adjacent([1,2,3,10,12,31], close=3)
     grp -> [2, 11, 31]
     idx -> [[0,1,2], [3,4], [5]]

    :param values: array of values to be grouped
    :param close: float
    :return grouped_values: float array(n) of grouped values
    :return indexes: [n] list of lists, each item relates to an averaged group, with indexes from values
    """
    # Check distance between good peaks
    dist_chk = []
    dist_idx = []
    gx = 0
    dist = [values[gx]]
    idx = [gx]
    while gx < len(values) - 1:
        gx += 1
        if (values[gx] - values[gx - 1]) < close:
            dist += [values[gx]]
            idx += [gx]
        else:
            dist_chk += [np.mean(dist)]
            dist_idx += [idx]
            dist = [values[gx]]
            idx = [gx]
    dist_chk += [np.mean(dist)]
    dist_idx += [idx]
    return np.array(dist_chk), dist_idx


def local_maxima_1d(y: np.ndarray) -> np.ndarray:
    """
    Find local maxima in 1d array
    Returns points with central point higher than neighboring points.

    Copied from scipy.signal._peak_finding_utils
    https://github.com/scipy/scipy/blob/v1.7.1/scipy/signal/_peak_finding_utils.pyx

    :param y: list or array
    :return: array of peak indexes
    """
    y = np.asarray(y, dtype=float).reshape(-1)

    # Preallocate, there can't be more maxima than half the size of `y`
    midpoints = np.empty(y.shape[0] // 2, dtype=np.intp)
    m = 0  # Pointer to the end of valid area in allocated arrays
    i = 1  # Pointer to current sample, first one can't be maxima
    i_max = y.shape[0] - 1  # Last sample can't be maxima
    while i < i_max:
        # Test if previous sample is smaller
        if y[i - 1] < y[i]:
            i_ahead = i + 1  # Index to look ahead of current sample

            # Find next sample that is unequal to x[i]
            while i_ahead < i_max and y[i_ahead] == y[i]:
                i_ahead += 1

            # Maxima is found if next unequal sample is smaller than x[i]
            if y[i_ahead] < y[i]:
                left_edge = i
                right_edge = i_ahead - 1
                midpoints[m] = (left_edge + right_edge) // 2
                m += 1
                # Skip samples that can't be maximum
                i = i_ahead
        i += 1
    return midpoints[:m]


def find_local_maxima(y: np.ndarray, yerror: np.ndarray | None = None) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Find local maxima in 1d arrays, returns index of local maximums, plus
    estimation of the peak power for each maxima and a classification of whether the maxima is greater than
    the standard deviation of the error.

    E.G.
        index, power, isgood = find_local_maxima(ydata)
        maxima = ydata[index[isgood]]
        maxima_power = power[isgood]
    
    Peak Power:
      peak power for each maxima is calculated using the peak_ratio algorithm for each maxima and adjacent points
    Good Peaks:
      Maxima are returned Good if:  power > (max(y) - min(y)) / std(yerror)
    
    :param y: array(n) of data
    :param yerror: array(n) of errors on data, or None to use default error function (sqrt(abs(y)+1))
    :return index: array(m<n) of indexes in y of maxima
    :return power: array(m) of estimated peak power for each maxima
    :return isgood: bool array(m) where True elements have power > power of the array
    """

    if yerror is None or np.all(np.abs(yerror) < 0.1):
        yerror = poisson_errors(y)
    else:
        yerror = np.asarray(yerror, dtype=float)
    yerror[yerror < 1] = 1.0
    bkg = np.min(y)
    wi = 1 / yerror ** 2

    index = local_maxima_1d(y)
    # average nearest 3 points to peak
    power = np.array([np.sum(wi[m-1:m+2] * (y[m-1:m+2] - bkg)) / np.sum(wi[m-1:m+2]) for m in index])
    # Determine if peak is good
    isgood = power > (np.max(y) - np.min(y)) / (np.std(yerror) + 1)
    return index, power, isgood


def find_peaks(y: np.ndarray, yerror: np.ndarray | None = None, 
               min_peak_power: float | None = None, peak_distance_idx: int = 6) -> tuple[np.ndarray, np.ndarray]:
    """
    Find peak shaps in linear-spaced 1d arrays with poisson like numerical values
    
    E.G.
      index, power = find_peaks(ydata, yerror, min_peak_power=None, peak_distance_idx=10)
      peak_centres = xdata[index]  # ordered by peak strength

    :param y: array(n) of data
    :param yerror: array(n) of errors on data, or None to use default error function (sqrt(abs(y)+1))
    :param min_peak_power: float, only return peaks with power greater than this. If None compare against std(y)
    :param peak_distance_idx: int, group adjacent maxima if closer in index than this
    :return index: array(m) of indexes in y of peaks that satisfy conditions
    :return power: array(m) of estimated power of each peak
    """
    # Get all peak positions
    midpoints, peak_signals, chk = find_local_maxima(y, yerror)

    if min_peak_power is None:
        good_peaks = chk
    else:
        good_peaks = peak_signals >= min_peak_power

    # select indexes of good peaks
    peaks_idx = midpoints[good_peaks]
    peak_power = peak_signals[good_peaks]
    if len(peaks_idx) == 0:
        return peaks_idx, peak_power

    # Average peaks close to each other
    group_idx, group_signal_idx = group_adjacent(peaks_idx, peak_distance_idx)
    peaks_idx = np.round(group_idx).astype(int)
    peak_power = np.array([np.sum(peak_power[ii]) for ii in group_signal_idx])

    # sort peak order by strength
    power_sort = np.argsort(peak_power)[::-1]
    return peaks_idx[power_sort], peak_power[power_sort]


def find_peaks_str(x: np.ndarray, y: np.ndarray, yerror: np.ndarray | None = None,
                   min_peak_power: float | None = None, peak_distance_idx: int = 6) -> str:
    """
    Find peak shaps in linear-spaced 1d arrays with poisson like numerical values

    E.G.
      index, power = find_peaks(ydata, yerror, min_peak_power=None, peak_distance_idx=10)
      peak_centres = xdata[index]  # ordered by peak strength

    :param x: array(n) of data
    :param y: array(n) of data
    :param yerror: array(n) of errors on data, or None to use default error function (sqrt(abs(y)+1))
    :param min_peak_power: float, only return peaks with power greater than this. If None compare against std(y)
    :param peak_distance_idx: int, group adjacent maxima if closer in index than this
    :return index: array(m) of indexes in y of peaks that satisfy conditions
    :return power: array(m) of estimated power of each peak
    """
    index, power = find_peaks(y, yerror, min_peak_power, peak_distance_idx)
    x_vals = x[index]
    out = f"Find Peaks:\n len: {len(x)}, max: {np.max(y):.5g}, min: {np.min(y):.5g}\n\n"
    out += '\n'.join(f"  {idx:4} {_x:10.5}  power={pwr:.3}" for idx, _x, pwr in zip(index, x_vals, power))
    return out


def max_index(array: np.ndarray) -> tuple[int, ...]:
    """Return the index of the largest value in an array."""
    max_idx = np.nanargmax(array)
    return np.unravel_index(max_idx, array.shape)


def peak_results(res: ModelResult) -> dict:
    """
    Generate dict of fit results, including summed totals
    totals = peak_results(res)
    totals = {
        'lmfit': lmfit_result (res),
        'npeaks': number of peak models used,
        'chisqr': Chi^2 of fit,
        'xdata': x-data used for fit,
        'ydata': y-data used for fit,
        'yfit': y-fit values,
        'weights': res.weights,
        'yerror': 1 / res.weights if res.weights is not None else 0 * res.data,
        # plus data from components, e.g.
        'p1_amplitude': Peak 1 area,
        'p1_fwhm': Peak 1 full-width and half-maximum
        'p1_center': Peak 1 peak position
        'p1_height': Peak 1 fitted height,
        # plus data for total fit:
        'amplitude': Total summed area,
        'center': average centre of peaks,
        'height': average height of peaks,
        'fwhm': average FWHM of peaks,
        'background': fitted background,
        # plut the errors on all parameters, e.g.
        'stderr_amplitude': error on 'amplitude',
    }
    :param res: lmfit fit result - ModelResult
    :return: {totals: (value, error)}
    """
    peak_prefx = [mod.prefix for mod in res.components if 'bkg' not in mod.prefix]
    npeaks = len(peak_prefx)
    nn = 1 / len(peak_prefx) if len(peak_prefx) > 0 else 1  # normalise by number of peaks
    comps = res.eval_components()
    # TODO: standardise these keys
    fit_dict = {
        'lmfit': res,
        'npeaks': npeaks,
        'peak_prefixes': peak_prefx,
        'chisqr': res.chisqr,
        'xdata': res.userkws['x'],
        'ydata': res.data,
        'weights': res.weights,
        'yerror': 1 / res.weights if res.weights is not None else 0 * res.data,
        'yfit': res.best_fit,
    }
    for comp_prefx, comp in comps.items():
        fit_dict['%sfit' % comp_prefx] = comp
    for pname, param in res.params.items():
        ename = 'stderr_' + pname
        fit_dict[pname] = param.value
        fit_dict[ename] = param.stderr if param.stderr is not None else 0
    totals = {
        'amplitude': np.sum([res.params['%samplitude' % pfx].value for pfx in peak_prefx]),
        'center': np.mean([res.params['%scenter' % pfx].value for pfx in peak_prefx]),
        'sigma': np.mean([res.params['%ssigma' % pfx].value for pfx in peak_prefx]),
        'height': np.mean([res.params['%sheight' % pfx].value for pfx in peak_prefx]),
        'fwhm': np.mean([res.params['%sfwhm' % pfx].value for pfx in peak_prefx]),
        'background': np.mean(comps['bkg_']) if 'bkg_' in comps else 0.0,
        'stderr_amplitude': np.sqrt(np.sum([fit_dict['stderr_%samplitude' % pfx] ** 2 for pfx in peak_prefx])),
        'stderr_center': np.sqrt(np.sum([fit_dict['stderr_%scenter' % pfx] ** 2 for pfx in peak_prefx])) * nn,
        'stderr_sigma': np.sqrt(np.sum([fit_dict['stderr_%ssigma' % pfx] ** 2 for pfx in peak_prefx])) * nn,
        'stderr_height': np.sqrt(np.sum([fit_dict['stderr_%sheight' % pfx] ** 2 for pfx in peak_prefx])) * nn,
        'stderr_fwhm': np.sqrt(np.sum([fit_dict['stderr_%sfwhm' % pfx] ** 2 for pfx in peak_prefx])) * nn,
        'stderr_background': np.std(comps['bkg_']) if 'bkg_' in comps else 0.0,
    }
    fit_dict.update(totals)
    return fit_dict


def peak_results_str(res: ModelResult) -> str:
    """
    Generate output str from lmfit results, including totals
    :param res: lmfit_result
    :return: str
    """
    fit_dict = peak_results(res)
    out = 'Fit Results\n'
    out += '%s\n' % res.model.name
    out += 'Npeaks = %d\n' % fit_dict['npeaks']
    out += 'Method: %s => %s\n' % (res.method, res.message)
    out += 'Chisqr = %1.5g\n' % res.chisqr
    # Peaks
    peak_prefx = [mod.prefix for mod in res.components if 'bkg' not in mod.prefix]
    for prefx in peak_prefx:
        out += '\nPeak %s\n' % prefx
        for pn in res.params:
            if prefx in pn:
                out += '%15s = %s\n' % (pn, stfm(fit_dict[pn], fit_dict['stderr_%s' % pn]))

    out += '\nBackground\n'
    for pn in res.params:
        if 'bkg' in pn:
            out += '%15s = %s\n' % (pn, stfm(fit_dict[pn], fit_dict['stderr_%s' % pn]))

    out += '\nTotals\n'
    out += '      amplitude = %s\n' % stfm(fit_dict['amplitude'], fit_dict['stderr_amplitude'])
    out += '         center = %s\n' % stfm(fit_dict['center'], fit_dict['stderr_center'])
    out += '         height = %s\n' % stfm(fit_dict['height'], fit_dict['stderr_height'])
    out += '          sigma = %s\n' % stfm(fit_dict['sigma'], fit_dict['stderr_sigma'])
    out += '           fwhm = %s\n' % stfm(fit_dict['fwhm'], fit_dict['stderr_fwhm'])
    out += '     background = %s\n' % stfm(fit_dict['background'], fit_dict['stderr_background'])
    return out


def peak_results_fit(res: ModelResult, ntimes: int = 10, x_data: np.ndarray | None = None) -> tuple[np.ndarray, np.ndarray]:
    """
    Generate xfit, yfit data, interpolated to give smoother variation
    :param res: lmfit_result
    :param ntimes: int, number of points * old number of points
    :param x_data: x data to interpolate or None to use data from fit
    :return: xfit, yfit
    """
    old_x = res.userkws['x'] if x_data is None else x_data
    xfit = np.linspace(np.min(old_x), np.max(old_x), np.size(old_x) * ntimes)
    yfit = res.eval(x=xfit)
    return xfit, yfit


def peak_results_plot(res: ModelResult, axes=None, xlabel: str = None, ylabel: str = None, title: str = None):
    """
    Plot peak results
    :param res: lmfit result
    :param axes: None or matplotlib axes
    :param xlabel: None or str
    :param ylabel: None or str
    :param title: None or str
    :return: matplotlib figure or axes
    """
    xdata = res.userkws['x']
    if title is None:
        title = res.model.name

    if axes:
        ax = res.plot_fit(ax=axes, xlabel=xlabel, ylabel=ylabel)
        # Add peak components
        comps = res.eval_components(x=xdata)
        for component in comps.keys():
            ax.plot(xdata, comps[component], label=component)
            ax.legend()
        return ax

    fig = res.plot(xlabel=xlabel, ylabel=ylabel)
    try:
        fig, grid = fig  # Old version of LMFit
    except TypeError:
        pass

    ax1, ax2 = fig.axes
    ax1.set_title(title, wrap=True)
    # Add peak components
    comps = res.eval_components(x=xdata)
    for component in comps.keys():
        ax2.plot(xdata, comps[component], label=component)
        ax2.legend()
    fig.set_figwidth(8)
    # fig.show()
    return fig


class FitResults:
    """
    FitResults Class
    Wrapper for lmfit results object with additional functions specific to i16_peakfit

    res = model.fit(ydata, x=xdata)  # lmfit ouput
    fitres = FitResults(res)

    --- Parameters ---
    fitres.res  # lmfit output
    # data from fit:
    fitres.npeaks # number of peak models used,
    fitres.chisqr  # Chi^2 of fit,
    fitres.xdata  # x-data used for fit,
    fitres.ydata  # y-data used for fit,
    fitres.yfit  # y-fit values,
    fitres.weights  # res.weights,
    fitres.yerror  # 1 / res.weights if res.weights is not None else 0 * res.data,
    # data from components, e.g.
    fitres.p1_amplitude  # Peak 1 area,
    fitres.p1_fwhm  # Peak 1 full-width and half-maximum
    fitres.p1_center  # Peak 1 peak position
    fitres.p1_height  # Peak 1 fitted height,
    # data for total fit:
    fitres.amplitude  # Total summed area,
    fitres.center  # average centre of peaks,
    fitres.height  # average height of peaks,
    fitres.fwhm  # average FWHM of peaks,
    fitres.background  # fitted background,
    # errors on all parameters, e.g.
    fitres.stderr_amplitude  # error on 'amplitude

    --- Functions ---
    print(fitres)  # prints formatted str with results
    ouputdict = fitres.results()  # creates ouput dict
    xdata, yfit = fitres.fit(ntimes=10)  # interpolated fit results
    fig = fitres.plot(axes, xlabel, ylabel, title)  # create plot
    """
    npeaks: int
    peak_prefixes: list[str]
    amplitude: float
    center: float
    height: float
    fwhm: float
    background: float
    stderr_amplitude: float
    stderr_center: float
    stderr_height: float
    stderr_fwhm: float
    stderr_background: float

    def __init__(self, results: ModelResult):
        self.res = results
        self._res = peak_results(results)
        for name in self._res:
            setattr(self, name, self._res[name])

    def __str__(self):
        return peak_results_str(self.res)

    def get_value(self, name: str) -> tuple[float | None, float]:
        """Returns fit parameter value and associated error"""
        err_name = f"stderr_{name}"
        value = self._res.get(name, None)
        error = self._res.get(err_name, 0)
        return value, error

    def get_string(self, name: str) -> str:
        """Returns fit parameter string including error in standard form"""
        value, error = self.get_value(name)
        return stfm(value, error)

    def results(self) -> dict:
        """Returns dict of peak fit results"""
        return self._res

    def fit_data(self, x_data: np.ndarray | None = None, ntimes=10) -> tuple[np.ndarray, np.ndarray]:
        """Returns interpolated x, y fit arrays"""
        return peak_results_fit(self.res, ntimes=ntimes, x_data=x_data)

    def plot(self, axes=None, xlabel=None, ylabel=None, title=None):
        """Plot peak fit results"""
        return peak_results_plot(self.res, axes, xlabel, ylabel, title)


def modelfit(xvals: np.ndarray, yvals: np.ndarray, yerrors: np.ndarray | None =None, 
             model=None | Model, initial_parameters: dict | None = None, 
             fix_parameters: dict | None = None, method: str = 'leastsq', 
             print_result: bool = False, plot_result: bool = False) -> ModelResult:
    """
    Fit x,y data to a model from lmfit
    E.G.:
      res = modelfit(x, y, model='Gauss')
      print(res.fit_report())
      res.plot()
      val = res.params['amplitude'].value
      err = res.params['amplitude'].stderr

    Model:
     from lmfit import models
     model1 = model.GaussianModel()
     model2 = model.LinearModel()
     model = model1 + model2
     res = model.fit(y, x=x)

    Provide initial guess:
      res = modelfit(x, y, model=VoightModel(), initial_parameters={'center':1.23})

    Fix parameter:
      res = modelfit(x, y, model=VoightModel(), fix_parameters={'sigma': fwhm/2.3548200})

    :param xvals: array(n) position data
    :param yvals: array(n) intensity data
    :param yerrors: None or array(n) - error data to pass to fitting function as weights: 1/errors^2
    :param model: lmfit.Model
    :param initial_parameters: None or dict of initial values for parameters
    :param fix_parameters: None or dict of parameters to fix at positions
    :param method: str method name, from lmfit fitting methods
    :param print_result: if True, prints the fit results using fit.fit_report()
    :param plot_result: if True, plots the results using fit.plot()
    :return: lmfit.model.ModelResult < fit results object
    """

    xvals = np.asarray(xvals, dtype=float).reshape(-1)
    yvals = np.asarray(yvals, dtype=float).reshape(-1)
    weights = gen_weights(yerrors)

    if initial_parameters is None:
        initial_parameters = {}
    if fix_parameters is None:
        fix_parameters = {}

    if model is None:
        model = GaussianModel() + LinearModel()

    pars = model.make_params()

    # user input parameters
    for ipar, ival in initial_parameters.items():
        if ipar in pars:
            pars[ipar].set(value=ival, vary=True)
    for ipar, ival in fix_parameters.items():
        if ipar in pars:
            pars[ipar].set(value=ival, vary=False)

    res = model.fit(yvals, pars, x=xvals, weights=weights, method=method)

    if print_result:
        print(res.fit_report())
    if plot_result:
        res.plot()
    return res


def peakfit(xvals: np.ndarray, yvals: np.ndarray, yerrors: np.ndarray | None = None, 
            model: str = 'Voight', background: str = 'slope',
            initial_parameters: dict | None = None, fix_parameters: dict | None = None, 
            method: str = 'leastsq', print_result: bool = False, plot_result: bool = False) -> FitResults:
    """
    Fit x,y data to a peak model using lmfit

    E.G.:
      res = peakfit(x, y, model='Gauss')
      print(res)
      res.plot()
      val = res.params['amplitude'].value
      err = res.params['amplitude'].stderr

    Peak Models:
     Choice of peak model: 'Gaussian', 'Lorentzian', 'Voight',' PseudoVoight'
    Background Models:
     Choice of background model: 'slope', 'exponential'

    Peak Parameters:
     'amplitude', 'center', 'sigma', pvoight only: 'fraction'
     output only: 'fwhm', 'height'
    Background parameters:
     'bkg_slope', 'bkg_intercept', or for exponential: 'bkg_amplitude', 'bkg_decay'

    Provide initial guess:
      res = peakfit(x, y, model='Voight', initial_parameters={'center': 1.23})

    Fix parameter:
      res = peakfit(x, y, model='gauss', fix_parameters={'sigma': fwhm/2.3548200})

    :param xvals: array(n) position data
    :param yvals: array(n) intensity data
    :param yerrors: None or array(n) - error data to pass to fitting function as weights: 1/errors^2
    :param model: str, specify the peak model: 'Gaussian','Lorentzian','Voight'
    :param background: str, specify the background model: 'slope', 'exponential'
    :param initial_parameters: None or dict of initial values for parameters
    :param fix_parameters: None or dict of parameters to fix at positions
    :param method: str method name, from lmfit fitting methods
    :param print_result: if True, prints the fit results using fit.fit_report()
    :param plot_result: if True, plots the results using fit.plot()
    :return: fit result object
    """

    xvals = np.asarray(xvals, dtype=float).reshape(-1)
    yvals = np.asarray(yvals, dtype=float).reshape(-1)
    weights = gen_weights(yerrors)

    if initial_parameters is None:
        initial_parameters = {}
    if fix_parameters is None:
        fix_parameters = {}

    peak_mod = None
    bkg_mod = None
    for model_name, names in PEAK_MODELS.items():
        if model.lower() in names:
            peak_mod = MODELS[model_name]()
    for model_name, names in BACKGROUND_MODELS.items():
        if background.lower() in names:
            bkg_mod = MODELS[model_name](prefix='bkg_')

    pars = peak_mod.guess(yvals, x=xvals)
    pars += bkg_mod.make_params()
    # pars += bkg_mod.make_params(intercept=np.min(yvals), slope=0)
    # pars['gamma'].set(value=0.7, vary=True, expr='') # don't fix gamma

    # user input parameters
    for ipar, ival in initial_parameters.items():
        if ipar in pars:
            pars[ipar].set(value=ival, vary=True)
    for ipar, ival in fix_parameters.items():
        if ipar in pars:
            pars[ipar].set(value=ival, vary=False)

    mod = peak_mod + bkg_mod
    res = mod.fit(yvals, pars, x=xvals, weights=weights, method=method)

    if print_result:
        print(peak_results_str(res))
    if plot_result:
        peak_results_plot(res)
    return FitResults(res)


def peak2dfit(xdata: np.ndarray, ydata: np.ndarray, image_data: np.ndarray, 
              initial_parameters: dict | None = None, fix_parameters: dict | None = None,
              print_result: bool = False, plot_result: bool = False):
    """
    Fit Gaussian Peak in 2D
    *** requires lmfit > 1.0.3 ***
        Not yet finished!
    :param xdata:
    :param ydata:
    :param image_data:
    :param initial_parameters:
    :param fix_parameters:
    :param print_result:
    :param plot_result:
    :return:
    """
    print('Not yet finished...')
    pass


def generate_model(xvals: np.ndarray, yvals: np.ndarray, yerrors: np.ndarray | None = None,
                   npeaks: int | None = None, min_peak_power: float | None = None, peak_distance_idx: int = 6,
                   model: str = 'Gaussian', background: str = 'slope', 
                   initial_parameters: dict | None = None, fix_parameters: dict | None = None) -> tuple[Model, Parameters]:
    """
    Generate lmfit profile models
    See: https://lmfit.github.io/lmfit-py/builtin_models.html#example-3-fitting-multiple-peaks-and-using-prefixes
    E.G.:
      mod, pars = generate_model(x, y, npeaks=1, model='Gauss', backgroud='slope')

    Peak Search:
     The number of peaks and initial peak centers will be estimated using the find_peaks function. If npeaks is given,
     the largest npeaks will be used initially. 'min_peak_power' and 'peak_distance_idx' can be input to tailor the
     peak search results.
     If the peak search returns < npeaks, fitting parameters will initially choose npeaks equally distributed points

    Peak Models:
     Choice of peak model: 'Gaussian', 'Lorentzian', 'Voight',' PseudoVoight'
    Background Models:
     Choice of background model: 'slope', 'exponential'

    :param xvals: array(n) position data
    :param yvals: array(n) intensity data
    :param yerrors: None or array(n) - error data to pass to fitting function as weights: 1/errors^2
    :param npeaks: None or int number of peaks to fit. None will guess the number of peaks
    :param min_peak_power: float, only return peaks with power greater than this. If None compare against std(y)
    :param peak_distance_idx: int, group adjacent maxima if closer in index than this
    :param model: str or lmfit.Model, specify the peak model 'Gaussian','Lorentzian','Voight'
    :param background: str, specify the background model: 'slope', 'exponential'
    :param initial_parameters: None or dict of initial values for parameters
    :param fix_parameters: None or dict of parameters to fix at positions
    :return: model, parameters
    """
    xvals = np.asarray(xvals, dtype=float).reshape(-1)
    yvals = np.asarray(yvals, dtype=float).reshape(-1)

    # Find peaks
    peak_idx, peak_pow = find_peaks(yvals, yerrors, min_peak_power, peak_distance_idx)
    peak_centers = {'p%d_center' % (n+1): xvals[peak_idx[n]] for n in range(len(peak_idx))}
    if npeaks is None:
        npeaks = len(peak_centers)

    if initial_parameters is None:
        initial_parameters = {}
    if fix_parameters is None:
        fix_parameters = {}

    peak_mod = None
    bkg_mod = None
    for model_name, names in PEAK_MODELS.items():
        if model.lower() in names:
            peak_mod = MODELS[model_name]
    for model_name, names in BACKGROUND_MODELS.items():
        if background.lower() in names:
            bkg_mod = MODELS[model_name]

    mod = bkg_mod(prefix='bkg_')
    for n in range(npeaks):
        mod += peak_mod(prefix='p%d_' % (n+1))

    pars = mod.make_params()

    # initial parameters
    min_wid = np.mean(np.diff(xvals))
    max_wid = xvals.max() - xvals.min()
    area = (yvals.max() - yvals.min()) * (3 * min_wid)
    percentile = np.linspace(0, 100, npeaks + 2)
    for n in range(1, npeaks+1):
        pars['p%d_amplitude' % n].set(value=area/npeaks, min=0)
        pars['p%d_sigma' % n].set(value=3*min_wid, min=min_wid, max=max_wid)
        pars['p%d_center' % n].set(value=np.percentile(xvals, percentile[n]), min=xvals.min(), max=xvals.max())
    # find_peak centers
    for ipar, ival in peak_centers.items():
        if ipar in pars:
            pars[ipar].set(value=ival, vary=True)
    # user input parameters
    for ipar, ival in initial_parameters.items():
        if ipar in pars:
            pars[ipar].set(value=ival, vary=True)
    for ipar, ival in fix_parameters.items():
        if ipar in pars:
            pars[ipar].set(value=ival, vary=False)
    return mod, pars


def generate_model_script(xvals: np.ndarray, yvals: np.ndarray, yerrors: np.ndarray = None,
                          npeaks: int | None = None, min_peak_power: float | None = None, peak_distance_idx: int = 6,
                          model: str = 'Gaussian', background: str = 'slope', 
                          initial_parameters: dict | None = None, fix_parameters: dict | None = None,
                          only_lmfit: bool = False) -> str:
    """
    Generate script to create lmfit profile models
    E.G.:
      string = generate_mode_stringl(x, y, npeaks=1, model='Gauss', backgroud='slope')

    :param xvals: array(n) position data
    :param yvals: array(n) intensity data
    :param yerrors: None or array(n) - error data to pass to fitting function as weights: 1/errors^2
    :param npeaks: None or int number of peaks to fit. None will guess the number of peaks
    :param min_peak_power: float, only return peaks with power greater than this. If None compare against std(y)
    :param peak_distance_idx: int, group adjacent maxima if closer in index than this
    :param model: str or lmfit.Model, specify the peak model 'Gaussian','Lorentzian','Voight'
    :param background: str, specify the background model: 'slope', 'exponential'
    :param initial_parameters: None or dict of initial values for parameters
    :param fix_parameters: None or dict of parameters to fix at positions
    :param only_lmfit: if True, only include lmfit imports
    :return: str
    """

    data = "xdata = np.array(%s)\n" % list(xvals)
    data += "ydata = np.array(%s)\n" % list(yvals)
    if yerrors is None or np.all(np.abs(yerrors) < 0.001):
        data += 'yerrors = None\n'
        data += 'weights = None\n\n'
    else:
        data += "yerrors = np.array(%s)\n" % list(yerrors)
        data += "yerrors[yerrors < 1] = 1.0\n"
        data += "weights = 1 / yerrors\n\n"

    if initial_parameters is None:
        initial_parameters = {}
    if fix_parameters is None:
        fix_parameters = {}
    params = "initial = %s\nfixed = %s\n" % (initial_parameters, fix_parameters)

    if not only_lmfit:
        out = "import numpy as np\nfrom mmg_toolbox import fitting\n\n"
        out += data
        out += '%s\n' % params
        out += "mod, pars = fitting.generate_model(xdata, ydata, yerrors,\n" \
               "                                   npeaks=%s, min_peak_power=%s, peak_distance_idx=%s,\n" \
               "                                   model='%s', background='%s',\n" \
               "                                   initial_parameters=initial, fix_parameters=fixed)\n" % (
                   npeaks, min_peak_power, peak_distance_idx, model, background
               )
    else:
        # Find peaks
        peak_idx, peak_pow = find_peaks(yvals, yerrors, min_peak_power, peak_distance_idx)
        peak_centers = {'p%d_center' % (n + 1): xvals[peak_idx[n]] for n in range(len(peak_idx))}
        for model_name, names in PEAK_MODELS.items():
            if model.lower() in names:
                peak_mod = MODELS[model_name]
        for model_name, names in BACKGROUND_MODELS.items():
            if background.lower() in names:
                bkg_mod = MODELS[model_name]
        peak_name = peak_mod.__name__
        bkg_name = bkg_mod.__name__

        out = "import numpy as np\nfrom lmfit import models\n\n"
        out += data
        out += "%speak_centers = %s\n\n" % (params, peak_centers)
        out += "mod = models.%s(prefix='bkg_')\n" % bkg_name
        out += "for n in range(len(peak_centers)):\n    mod += models.%s(prefix='p%%d_' %% (n+1))\n" % peak_name
        out += "pars = mod.make_params()\n\n"
        out += "# initial parameters\n"
        out += "min_wid = np.mean(np.diff(xdata))\n"
        out += "max_wid = xdata.max() - xdata.min()\n"
        out += "area = (ydata.max() - ydata.min()) * (3 * min_wid)\n"
        out += "for n in range(1, len(peak_centers)+1):\n"
        out += "    pars['p%d_amplitude' % n].set(value=area/len(peak_centers), min=0)\n"
        out += "    pars['p%d_sigma' % n].set(value=3*min_wid, min=min_wid, max=max_wid)\n"
        out += "# find_peak centers\n"
        out += "for ipar, ival in peak_centers.items():\n"
        out += "    if ipar in pars:\n"
        out += "        pars[ipar].set(value=ival, vary=True)\n"
        out += "# user input parameters\n"
        out += "for ipar, ival in initial.items():\n"
        out += "    if ipar in pars:\n"
        out += "        pars[ipar].set(value=ival, vary=True)\n"
        out += "for ipar, ival in fixed.items():\n"
        out += "    if ipar in pars:\n"
        out += "        pars[ipar].set(value=ival, vary=False)\n\n"
    out += "# Fit data\n"
    out += "res = mod.fit(ydata, pars, x=xdata, weights=weights, method='leastsqr')\n"
    out += "print(res.fit_report())\n\n"
    out += "fig, grid = res.plot()\n"
    out += "ax1, ax2 = fig.axes\n"
    out += "comps = res.eval_components()\n"
    out += "for component in comps.keys():\n"
    out += "    ax2.plot(xdata, comps[component], mode=component)\n"
    out += "    ax2.legend()\n\n"
    return out


def multipeakfit(xvals: np.ndarray, yvals: np.ndarray, yerrors: np.ndarray | None = None,
                 npeaks: int | None = None, min_peak_power: float | None = None, 
                 peak_distance_idx: int | None = 10, model: str = 'Gaussian', background: str = 'slope', 
                 initial_parameters: dict | None = None, fix_parameters: dict | None = None, 
                 method: str ='leastsq', print_result: bool = False, plot_result: bool = False) -> FitResults:
    """
    Fit x,y data to a model with multiple peaks using lmfit
    See: https://lmfit.github.io/lmfit-py/builtin_models.html#example-3-fitting-multiple-peaks-and-using-prefixes
    
    E.G.:
      res = multipeakfit(x, y, npeaks=None, model='Gauss', plot_result=True)
      val = res.params['p1_amplitude'].value
      err = res.params['p1_amplitude'].stderr

    Peak Search:
     The number of peaks and initial peak centers will be estimated using the find_peaks function. If npeaks is given,
     the largest npeaks will be used initially. 'min_peak_power' and 'peak_distance_idx' can be input to tailor the
     peak search results.
     If the peak search returns < npeaks, fitting parameters will initially choose npeaks equally distributed points

    Peak Models:
     Choice of peak model: 'Gaussian', 'Lorentzian', 'Voight',' PseudoVoight'
    Background Models:
     Choice of background model: 'slope', 'exponential'

    Peak Parameters (%d=number of peak):
    Parameters in '.._parameters' dicts and in output results. Each peak (upto npeaks) has a set number of parameters:
     'p%d_amplitude', 'p%d_center', 'p%d_dsigma', pvoight only: 'p%d_fraction'
     output only: 'p%d_fwhm', 'p%d_height'
    Background parameters:
     'bkg_slope', 'bkg_intercept', or for exponential: 'bkg_amplitude', 'bkg_decay'

    Provide initial guess:
      res = multipeakfit(x, y, model='Voight', initial_parameters={'p1_center':1.23})

    Fix parameter:
      res = multipeakfit(x, y, model='gauss', fix_parameters={'p1_sigma': fwhm/2.3548200})

    :param xvals: array(n) position data
    :param yvals: array(n) intensity data
    :param yerrors: None or array(n) - error data to pass to fitting function as weights: 1/errors^2
    :param npeaks: None or int number of peaks to fit. None will guess the number of peaks
    :param min_peak_power: float, only return peaks with power greater than this. If None compare against std(y)
    :param peak_distance_idx: int, group adjacent maxima if closer in index than this
    :param model: str or lmfit.Model, specify the peak model 'Gaussian','Lorentzian','Voight'
    :param background: str, specify the background model: 'slope', 'exponential'
    :param initial_parameters: None or dict of initial values for parameters
    :param fix_parameters: None or dict of parameters to fix at positions
    :param method: str method name, from lmfit fitting methods
    :param print_result: if True, prints the fit results using fit.fit_report()
    :param plot_result: if True, plots the results using fit.plot()
    :return: FitResults object
    """
    xvals = np.asarray(xvals, dtype=float).reshape(-1)
    yvals = np.asarray(yvals, dtype=float).reshape(-1)
    weights = gen_weights(yerrors)

    mod, pars = generate_model(xvals, yvals, yerrors,
                               npeaks=npeaks, min_peak_power=min_peak_power, peak_distance_idx=peak_distance_idx,
                               model=model, background=background,
                               initial_parameters=initial_parameters, fix_parameters=fix_parameters)

    # Fit data against model using choosen method
    res = mod.fit(yvals, pars, x=xvals, weights=weights, method=method)

    if print_result:
        print(peak_results_str(res))
    if plot_result:
        peak_results_plot(res)
    return FitResults(res)


"----------------------------------------------------------------------------------------------------------------------"
"------------------------------------------------ ScanFitManager ------------------------------------------------------"
"----------------------------------------------------------------------------------------------------------------------"


class ScanFitManager:
    """
    ScanFitManager
     Holds several functions for automatically fitting scan data

    fit = ScanFitManager(scan)
    fit.peak_ratio(yaxis)  # calculates peak power
    fit.find_peaks(xaxis, yaxis)  # automated peak finding routine
    fit.fit(xaxis, yaxis)  # estimate & fit data against a peak profile model using lmfit
    fit.multi_peak_fit(xaxis, yaxis)  # find peaks & fit multiprofile model using lmfit
    fit.model_fit(xaxis, yaxis, model, pars)  # fit supplied model against data
    fit.fit_results()  # return lmfit.ModelResult for last fit
    fit.fit_values()  # return dict of fit values for last fit
    fit.fit_report()  # return str of fit report
    fit.plot()  # plot last lmfit results
    * xaxis, yaxis are str names of arrays in the scan namespace

    :param scan: babelscan.Scan
    """

    def __init__(self, scan: NexusScan):
        self.scan = scan

    def __call__(self, *args, **kwargs) -> FitResults:
        """Calls ScanFitManager.fit(...)"""
        return self.multi_peak_fit(*args, **kwargs)

    def __str__(self):
        return self.fit_report()

    def peak_ratio(self, yaxis: str = 'signal') -> float:
        """
        Return the ratio signal / error for given dataset
        From Blessing, J. Appl. Cryst. (1997). 30, 421-426 Equ: (1) + (6)
          peak_ratio = (sum((y-bkg)/dy^2)/sum(1/dy^2)) / sqrt(i/sum(1/dy^2))
        :param yaxis: str name or address of array to plot on y axis
        :return: float ratio signal / err
        """
        values = self.scan.eval(yaxis)
        errors = np.sqrt(values + 0.1)
        return peak_ratio(values, errors)

    def find_peaks(self, xaxis: str = 'axes', yaxis: str = 'signal', min_peak_power: float | None = None, 
                   peak_distance_idx: int = 6) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Find peak shaps in linear-spaced 1d arrays with poisson like numerical values
        
        E.G.
          centres, index, power = self.find_peaks(xaxis, yaxis, min_peak_power=None, peak_distance_idx=10)
        
        :param xaxis: str name or address of array to plot on x axis
        :param yaxis: str name or address of array to plot on y axis
        :param min_peak_power: float, only return peaks with power greater than this. If None compare against std(y)
        :param peak_distance_idx: int, group adjacent maxima if closer in index than this
        :return centres: array(m) of peak centers in x, equiv. to xdata[index]
        :return index: array(m) of indexes in y of peaks that satisfy conditions
        :return power: array(m) of estimated power of each peak
        """
        xdata, ydata = self.scan.arrays(xaxis, yaxis)
        errors = np.sqrt(ydata + 0.1)
        index, power = find_peaks(ydata, errors, min_peak_power, peak_distance_idx)
        return xdata[index], index, power

    def fit(self, xaxis: str = 'axes', yaxis: str = 'signal', model: str = 'Gaussian', 
            background: str = 'slope', initial_parameters: dict | None = None, 
            fix_parameters: dict | None = None, method: str = 'leastsq', 
            print_result: bool = False, plot_result: bool = False) -> FitResults:
        """
        Fit x,y data to a peak model using lmfit

        E.G.:
          res = self.fit('axes', 'signal', model='Gauss')
          print(res)
          res.plot()
          val1 = res.p1_amplitude
          val2 = res.p2_amplitude

        Peak Models:
         Choice of peak model: 'Gaussian', 'Lorentzian', 'Voight', 'PseudoVoight'
        Background Models:
         Choice of background model: 'flat', 'slope', 'exponential'

        Peak Parameters (%d=number of peak):
         'amplitude', 'center', 'sigma', pvoight only: 'fraction'
         output only: 'fwhm', 'height'
        Background parameters:
         'bkg_slope', 'bkg_intercept', or for exponential: 'bkg_amplitude', 'bkg_decay'
         output only: 'background'
        Uncertainties (errors):
         'stderr_PARAMETER', e.g. 'stderr_amplitude'

        Provide initial guess:
          res = self.fit(x, y, model='Voight', initial_parameters={'p1_center':1.23})

        Fix parameter:
          res = self.fit(x, y, model='gauss', fix_parameters={'p1_sigma': fwhm/2.3548200})

        :param xaxis: str name or address of array to plot on x axis
        :param yaxis: str name or address of array to plot on y axis
        :param model: str, specify the peak model 'Gaussian','Lorentzian','Voight'
        :param background: str, specify the background model: 'slope', 'exponential'
        :param initial_parameters: None or dict of initial values for parameters
        :param fix_parameters: None or dict of parameters to fix at positions
        :param method: str method name, from lmfit fitting methods
        :param print_result: if True, prints the fit results using fit.fit_report()
        :param plot_result: if True, plots the results using fit.plot()
        :return: FitResult object
        """
        xdata, ydata = self.scan.arrays(xaxis, yaxis)
        errors = self.scan._error_function(ydata)
        xname, yname = self.scan.labels(xaxis, yaxis)

        # lmfit
        res = peakfit(xdata, ydata, errors, model=model, background=background,
                      initial_parameters=initial_parameters, fix_parameters=fix_parameters, method=method)

        output = res.results()
        output.update({
            'lmfit': res.res,
            'fit_result': res.res,
            'fitobj': res,
            'fit': res.res.best_fit,
            f"fit_{yname}": res.res.best_fit,
        })
        self.scan.map.add_local(**output)

        if print_result:
            print(self.scan.title())
            print(res)
        if plot_result:
            res.plot(title=self.scan.title())
        return res

    def multi_peak_fit(self, xaxis: str = 'axes', yaxis: str = 'signal',
                       npeaks: int | None = None, min_peak_power: int | None = None, peak_distance_idx: int = 6,
                       model: str = 'Gaussian', background: str = 'slope',
                       initial_parameters: dict | None = None, fix_parameters: dict | None = None, 
                       method: str = 'leastsq', print_result: bool = False, plot_result: bool = False
                       ) -> FitResults:
        """
        Fit x,y data to a peak model using lmfit

        E.G.:
          res = self.multi_peak_fit('axes', 'signal', npeaks=2, model='Gauss')
          print(res)
          res.plot()
          val1 = res.p1_amplitude
          val2 = res.p2_amplitude

        Peak centers:
         Will attempt a fit using 'npeaks' peaks, with centers defined by defalult by the find_peaks function
          if 'npeaks' is None, the number of peaks will be found by find_peaks()
          if 'npeaks' is greater than the number of peaks found by find_peaks, initial peak centers are evenly
          distrubuted along xdata.

        Peak Models:
         Choice of peak model: 'Gaussian', 'Lorentzian', 'Voight','PseudoVoight'
        Background Models:
         Choice of background model: 'flat', 'slope', 'exponential'

        Peak Parameters (%d=number of peak):
         'p%d_amplitude', 'p%d_center', 'p%d_sigma', pvoight only: 'p%d_fraction'
         output only: 'p%d_fwhm', 'p%d_height'
        Background parameters:
         'bkg_slope', 'bkg_intercept', or for exponential: 'bkg_amplitude', 'bkg_decay'
        Total parameters (always available, output only - sum/averages of all peaks):
         'amplitude', 'center', 'sigma', 'fwhm', 'height', 'background'
        Uncertainties (errors):
         'stderr_PARAMETER', e.g. 'stderr_amplitude'

        Provide initial guess:
          res = self.multi_peak_fit(x, y, model='Voight', initial_parameters={'p1_center':1.23})

        Fix parameter:
          res = self.multi_peak_fit(x, y, model='gauss', fix_parameters={'p1_sigma': fwhm/2.3548200})

        :param xaxis: str name or address of array to plot on x axis
        :param yaxis: str name or address of array to plot on y axis
        :param npeaks: None or int number of peaks to fit. None will guess the number of peaks
        :param min_peak_power: float, only return peaks with power greater than this. If None compare against std(y)
        :param peak_distance_idx: int, group adjacent maxima if closer in index than this
        :param model: str, specify the peak model 'Gaussian','Lorentzian','Voight'
        :param background: str, specify the background model: 'slope', 'exponential'
        :param initial_parameters: None or dict of initial values for parameters
        :param fix_parameters: None or dict of parameters to fix at positions
        :param method: str method name, from lmfit fitting methods
        :param print_result: if True, prints the fit results using fit.fit_report()
        :param plot_result: if True, plots the results using fit.plot()
        :return: FitResults object
        """
        xdata, ydata = self.scan.eval(f"{xaxis}, {yaxis}", default=np.zeros(self.scan.map.scannables_length()))
        errors = np.sqrt(ydata + 0.1)
        xname, yname = self.scan.map.generate_ids(xaxis, yaxis)

        # lmfit
        res = multipeakfit(xdata, ydata, errors, npeaks=npeaks, min_peak_power=min_peak_power,
                           peak_distance_idx=peak_distance_idx, model=model, background=background,
                           initial_parameters=initial_parameters, fix_parameters=fix_parameters, method=method)

        output = res.results()
        output.update({
            'lmfit': res.res,
            'fit_result': res.res,
            'fitobj': res,
            'fit': res.res.best_fit,
            f"fit_{yname}": res.res.best_fit,
        })
        self.scan.map.add_local(**output)

        if print_result:
            print(self.scan.title())
            print(res)
        if plot_result:
            res.plot(title=self.scan.title())
        return res

    def modelfit(self, xaxis: str = 'axis', yaxis: str = 'signal', model: Model | None = None, 
                 pars: Parameters | None = None, method: str = 'leastsq',
                 print_result: bool = False, plot_result: bool = False) -> ModelResult:
        """
        Fit data from scan against lmfit model

        Example:
            from lmfit.models import GaussianModel, LinearModel
            mod = GaussainModel(prefix='p1_') + LinearModel(prefix='bkg_')
            pars = mod.make_params()
            pars['p1_center'].set(value=np.mean(x), min=x.min(), max=x.max())
            res = scan.fit.modelfit('axis', 'signal', mod, pars)
            print(res.fit_report())
            res.plot()
            area = res.params['p1_amplitude'].value
            err = res.params['p1_amplitude'].stderr

        :param xaxis: str name or address of array to plot on x axis
        :param yaxis: str name or address of array to plot on y axis
        :param model: lmfit.Model - object defining combination of models
        :param pars: lmfit.Parameters - object defining model parameters
        :param method: str name of fitting method to use
        :param print_result: bool, if True, print results.fit_report()
        :param plot_result: bool, if True, generate results.plot()
        :return: lmfit fit results
        """
        data = self.scan.get_plot_data(xaxis, yaxis)
        xdata, ydata, yerror, xname, yname = (data.get(k) for k in ['x', 'y', 'yerror', 'xlabel', 'ylabel'])

        # weights
        if yerror is None or np.all(np.abs(yerror) < 0.001):
            weights = None
        else:
            weights = 1 / np.square(yerror, dtype=float)
            weights = np.nan_to_num(weights)

        # Default model, pars
        if model is None:
            model = LinearModel()
        if pars is None:
            pars = model.guess(ydata, x=xdata)

        # lmfit
        res = model.fit(ydata, pars, x=xdata, weights=weights, method=method)

        fit_dict = {
            'lmfit': res,
            'fit_result': res,
            'fit': res.best_fit,
            f"fit_{yname}": res.best_fit,
        }
        for pname, param in res.params.items():
            ename = 'stderr_' + pname
            fit_dict[pname] = param.value
            fit_dict[ename] = param.stderr

        # Add peak components
        comps = res.eval_components(x=xdata)
        for component in comps.keys():
            fit_dict[f"{component}fit"] = comps[component]
        self.scan.map.add_local(**fit_dict)

        if print_result:
            print(self.scan.title())
            print(res.fit_report())
        if plot_result:
            fig = res.plot(xlabel=xname, ylabel=yname)
            try:
                fig, grid = fig  # Old version of LMFit
            except TypeError:
                pass
            ax1, ax2 = fig.axes
            ax1.set_title(self.scan.title(), wrap=True)
        return res

    def gen_model(self, xaxis: str = 'axes', yaxis: str = 'signal',
                  npeaks: int | None = None, min_peak_power: float | None = None, 
                  peak_distance_idx: int = 6, model: str = 'Gaussian', background: str = 'slope',
                  initial_parameters: dict | None = None, fix_parameters: dict | None = None
                  ) -> tuple[Model, Parameters]:
        """
        Generate lmfit model and parameters

        :param xaxis: str name or address of array to plot on x axis
        :param yaxis: str name or address of array to plot on y axis
        :param npeaks: None or int number of peaks to fit. None will guess the number of peaks
        :param min_peak_power: float, only return peaks with power greater than this. If None compare against std(y)
        :param peak_distance_idx: int, group adjacent maxima if closer in index than this
        :param model: str, specify the peak model 'Gaussian','Lorentzian','Voight'
        :param background: str, specify the background model: 'slope', 'exponential'
        :param initial_parameters: None or dict of initial values for parameters
        :param fix_parameters: None or dict of parameters to fix at positions
        :returns: model, pars
        """
        data = self.scan.get_plot_data(xaxis, yaxis)
        xdata, ydata = (data[k] for k in ['x', 'y'])
        mod, pars = generate_model(xdata, ydata,
                                   npeaks=npeaks, min_peak_power=min_peak_power, peak_distance_idx=peak_distance_idx,
                                   model=model, background=background,
                                   initial_parameters=initial_parameters, fix_parameters=fix_parameters)
        return mod, pars

    def gen_model_script(self, xaxis: str = 'axes', yaxis: str = 'signal',
                         npeaks: int | None = None, min_peak_power: float | None = None, peak_distance_idx: int = 6,
                         model: str = 'Gaussian', background: str = 'slope',
                         initial_parameters: dict | None = None, fix_parameters: dict | None = None, 
                         only_lmfit: bool = False) -> str:
        """
        Generate script string of fit process
        :param xaxis: str name or address of array to plot on x axis
        :param yaxis: str name or address of array to plot on y axis
        :param npeaks: None or int number of peaks to fit. None will guess the number of peaks
        :param min_peak_power: float, only return peaks with power greater than this. If None compare against std(y)
        :param peak_distance_idx: int, group adjacent maxima if closer in index than this
        :param model: str, specify the peak model 'Gaussian','Lorentzian','Voight'
        :param background: str, specify the background model: 'slope', 'exponential'
        :param initial_parameters: None or dict of initial values for parameters
        :param fix_parameters: None or dict of parameters to fix at positions
        :param only_lmfit: if True, only include imports for lmfit
        :return: str
        """
        data = self.scan.get_plot_data(xaxis, yaxis)
        xdata, ydata, yerror = (data.get(k) for k in ['x', 'y', 'yerror'])
        out = generate_model_script(xdata, ydata, yerror,
                                    npeaks=npeaks, min_peak_power=min_peak_power,
                                    peak_distance_idx=peak_distance_idx,
                                    model=model, background=background,
                                    initial_parameters=initial_parameters, fix_parameters=fix_parameters,
                                    only_lmfit=only_lmfit)
        return out

    def gen_lmfit_script(self, xaxis: str = 'axes', yaxis: str = 'signal',
                         npeaks: int | None = None, min_peak_power: float | None = None, peak_distance_idx: int = 6,
                         model: str = 'Gaussian', background: str = 'slope',
                         initial_parameters: dict | None = None, fix_parameters: dict | None = None, ):
        """
        Generate script string of fit process, using only lmfit

        :param xaxis: str name or address of array to plot on x axis
        :param yaxis: str name or address of array to plot on y axis
        :param npeaks: None or int number of peaks to fit. None will guess the number of peaks
        :param min_peak_power: float, only return peaks with power greater than this. If None compare against std(y)
        :param peak_distance_idx: int, group adjacent maxima if closer in index than this
        :param model: str, specify the peak model 'Gaussian','Lorentzian','Voight'
        :param background: str, specify the background model: 'slope', 'exponential'
        :param initial_parameters: None or dict of initial values for parameters
        :param fix_parameters: None or dict of parameters to fix at positions
        :return: str
        """
        data = self.scan.get_plot_data(xaxis, yaxis)
        xdata, ydata, yerror = (data.get(k) for k in ['x', 'y', 'yerror'])
        out = generate_model_script(xdata, ydata, yerror,
                                    npeaks=npeaks, min_peak_power=min_peak_power,
                                    peak_distance_idx=peak_distance_idx,
                                    model=model, background=background,
                                    initial_parameters=initial_parameters, fix_parameters=fix_parameters,
                                    only_lmfit=True)
        return out

    def fit_parameter(self, parameter_name: str = 'amplitude') -> tuple[float, float]:
        """
        Returns parameter, error from the last run fit
        :param parameter_name: str, name from last fit e.g. 'amplitude', 'center', 'fwhm', 'background'
        :returns:  value, error
        """
        fitobj = self.fit_result()
        return fitobj.get_value(parameter_name)

    def fit_result(self) -> FitResults:
        """
        Returns FitResults object from last fit
        :return: PeakResults obect
        """
        return self.scan('fitobj')

    def fit_report(self) -> str:
        """Return str results of last fit"""
        fitobj = self.fit_result()
        return str(fitobj)

    def plot(self):
        """Plot fit results"""
        fitobj = self.fit_result()
        return fitobj.plot(title=self.scan.title())




"""
NeXus Scan Classes

NexusScan - NeXus Scan class, lazy loader of scan files
NexusDataHolder - Loads scan data and meta data into attributes
"""

import datetime

import h5py
import numpy as np
from hdfmap import NexusLoader, NexusMap, load_hdf
from hdfmap.eval_functions import dataset2data, dataset2str

from mmg_toolbox.beamline_metadata.hdfmap_generic import HdfMapMMGMetadata as Md
from mmg_toolbox.beamline_metadata.config import beamline_config, C
from mmg_toolbox.nexus.instrument_model import NXInstrumentModel
from mmg_toolbox.nexus.nexus_functions import get_dataset_value
from mmg_toolbox.utils.file_functions import get_scan_number
from mmg_toolbox.utils.misc_functions import shorten_string, DataHolder
from mmg_toolbox.xas import SpectraContainer, load_xas_scans


class NexusScan(NexusLoader):
    """
    Light-weight NeXus file reader

    Example:
        scan = NexusScan('scan.nxs')
        scan('scan_command') -> returns value

    :param nxs_filename: path to nexus file
    :param hdf_map: NexusMap object or None
    :param config: configuration dict
    """
    MAX_STR_LEN: int = 100

    def __init__(self, nxs_filename: str, hdf_map: NexusMap | None = None, config: dict | None = None):
        super().__init__(nxs_filename, hdf_map)
        self.config: dict = config or beamline_config()
        self.beamline = self.config.get('beamline', None)

        # add scan number to eval namespace
        self.map.add_local(scan_number=self.scan_number())

        from mmg_toolbox.utils.fitting import ScanFitManager, poisson_errors
        self.fit = ScanFitManager(self)
        self._error_function = poisson_errors
        from mmg_toolbox.plotting.scan_plot_manager import ScanPlotManager
        self.plot = ScanPlotManager(self)

    def __repr__(self):
        if self.beamline:
            return f"NexusScan<{self.beamline}>({self.scan_number()}: '{self.filename}')"
        return f"NexusScan('{self.filename}')"

    def __str__(self):
        try:
            return self.metadata_str()
        except Exception as ex:
            return f"{repr(self)}\n  Metadata failed with: \n{ex}\n"

    def metadata_str(self, expression: str | None = None):
        """Generate metadata string from beamline config"""
        if expression is None:
            expression = self.config.get(C.metadata_string, '')
        return self.format(expression)

    def scan_number(self) -> int:
        return get_scan_number(self.filename)

    def title(self) -> str:
        return f"#{self.scan_number()}"

    def label(self) -> str:
        return f"#{self.scan_number()}"

    def load_hdf(self) -> h5py.File:
        """Load the Hdf file"""
        return load_hdf(self.filename)

    def datasets(self, *args) -> list[h5py.Dataset]:
        """Return HDF5 datasets from NeXus file (leaves file in open state)"""
        with self.load_hdf() as hdf:
            return [hdf[self.map.combined[name]] for name in args]

    def arrays(self, *args, units: str = '', default: np.ndarray = np.array([np.nan])) -> list[np.ndarray]:
        """Return Numpy arrays"""
        with self.load_hdf() as hdf:
            return [
                get_dataset_value(self.map.combined[name], hdf, units=units, default=default)
                for name in args
            ]

    def values(self, *args, value_func=np.mean,
               units: str = '', default: np.ndarray = np.array(np.nan)) -> list[np.floating]:
        """Return float values"""
        with self.load_hdf() as hdf:
            return [
                value_func(get_dataset_value(self.map.combined[name], hdf, units=units, default=default))
                for name in args
            ]

    def times(self, *args) -> list[datetime.datetime]:
        """Return datetime object"""
        with self.load_hdf() as hdf:
            return [dataset2data(hdf[self.map.combined[name]]) for name in args]

    def strings(self, *args, units=False) -> list[str]:
        """Return string value"""
        with self.load_hdf() as hdf:
            return [dataset2str(hdf[self.map.combined[name]], units=units) for name in args]

    def image(self, index: int | tuple | slice | None = None) -> np.ndarray:
        """Return image or selection from default detector"""
        with self.load_hdf() as hdf:
            return self.map.get_image(hdf, index)

    def table(self, delimiter=', ', string_spec='', format_spec='f', default_decimals=8) -> str:
        """Return data table"""
        with self.load_hdf() as hdf:
            return self.map.create_scannables_table(hdf, delimiter, string_spec, format_spec, default_decimals)

    #TODO: Remove this?
    def get_plot_data(self, x_axis: str = 'axes0', y_axis: str = 'signal0') -> dict:
        with self.load_hdf() as hdf:
            cmd = self.map.eval(hdf, Md.cmd)
            if len(cmd) > self.MAX_STR_LEN:
                cmd = shorten_string(cmd)
            xdata = self.map.eval(hdf, x_axis)
            ydata = self.map.eval(hdf, y_axis)
            yerror = self._error_function(ydata)
            x_lab, y_lab = self.map.generate_ids(x_axis, y_axis, modify_missing=False)
            return {
                'x': xdata,
                'y': ydata,
                'yerror': yerror,
                'xlabel': x_lab,
                'ylabel': y_lab,
                'title': f"#{self.scan_number()}\n{cmd}"
            }

    def xas_scan(self) -> SpectraContainer:
        """Load XAS Spectra"""
        return load_xas_scans(self.filename)[0]

    def instrument_model(self) -> NXInstrumentModel:
        """return instrument model"""
        with self.load_hdf() as hdf:
            return NXInstrumentModel(hdf)


class NexusDataHolder(DataHolder, NexusScan):
    """
    Nexus data holder class
     - Automatically reads scannable and metadata from file
     - acts like the old .dat DataHolder class
     - has additional functions to read data from NeXus file

    Example:
        scan = NexusDataHolder('12345.nxs')
        scan.eta -> returns array
        scan.metadata.metadata -> returns value
        scan('signal') -> evaluate expression

    :param filename: path to Nexus file
    :param hdf_map: NexusMap object or None to generate
    :param flatten_scannables: if True, flattens all scannable arrays to 1D
    """
    filename: str
    map: NexusMap
    metadata: DataHolder

    def __init__(self, filename: str | None, hdf_map: NexusMap | None = None, flatten_scannables: bool = True,
                 config: dict | None = None):
        NexusScan.__init__(self, filename, hdf_map, config)

        with load_hdf(filename) as hdf:
            metadata = self.map.get_metadata(hdf)
            scannables = self.map.get_scannables(hdf, flatten=flatten_scannables)
        DataHolder.__init__(self, **scannables)
        self.metadata = DataHolder(**metadata)

    def __repr__(self):
        return f"NexusDataHolder('{self.filename}')"



"""
Create logging objects and functions

See https://docs.python.org/3/howto/logging.html
"""

import logging

# Setup config - doesn't work properly without this
logging.basicConfig()   # setup logging


def create_logger(name: str) -> logging.Logger:
    """Create new logger instance"""
    return logging.getLogger(name)


def set_all_logging_level(level: str | int):
    """
    Set logging level of all loggers
    Logging Levels (see builtin module logging)
        'notset'   |  0
        'debug'    |  10
        'info'     |  20
        'warning'  |  30
        'error'    |  40
        'critical' |  50
    :param level: str level name or int level
    :return: None
    """
    try:
        level = level.upper()
        # level = logging.getLevelNamesMapping()[level]  # Python >3.11
        level = logging._nameToLevel[level]
    except AttributeError:
        level = int(level)

    logging_logger = logging.getLogger(__name__)
    for logger in [logging.getLogger(name) for name in logging.root.manager.loggerDict]:
        logger.setLevel(level)
    logging_logger.info(f"Logging level set to {level}")


"""
Unit conversions
"""
import numpy as np

METERS = {  # conversion to meters
    'km': 1e3, 'm': 1, 'cm': 0.1, 'mm': 1e-3,
    'um': 1e-6, 'μm': 1e-6, 'nm': 1e-9,
    'a': 1e-10, 'A': 1e-10, 'Å': 1e-10, 'angstrom': 1e-10,
}

ENERGY = {  # conversion to eV
    'keV': 1e3, 'eV': 1, 'MeV': 1e6,
}

ANGLE = {  # conversion to degrees
    'deg': np.pi / 180, 'rad': np.pi / 180,
}


def unit_converter(value: float, from_unit: str, to_unit: str):
    """
    Converts a value from one unit to another
    """
    if from_unit in METERS:
        # Distances in meters
        distance = value * METERS[from_unit]
        return distance / METERS[to_unit]
    if from_unit in ENERGY:
        # Energy in eV
        energy = value * ENERGY[from_unit]
        return energy / ENERGY[to_unit]
    if from_unit in ANGLE:
        # Angle in degrees
        angle = value * ANGLE[from_unit]
        return angle / ANGLE[to_unit]
    raise ValueError(f"Unknown unit {from_unit}")


"""
Functions for communicating with GDA
"""

import stomp
import json
import sys

from .env_functions import get_beamline


CONTROL_NAME = "{beamline}-control"

def gda_datavis_file_message(filepath: str, connect_to: str | None = None):
    """
    Send an ActiveMQ message using stomp with a filepath.
    This will be picked up by any GDA instance connected to the control machine in the DataVis perspective.

    :param filepath: path to file to send, e.g. "/dls/i16/data/2022/cm31138-14/processed/960677_msmapper.nxs"
    :param connect_to: name of beamline control machine
    """
    if connect_to is None:
        beamline = get_beamline(filename=filepath)
        connect_to = CONTROL_NAME.format(beamline=beamline)

    conn = stomp.Connection([(connect_to, 61613)], auto_content_length=False)
    # conn.start()
    conn.connect()
    print(f"Connected to {connect_to}")

    message = json.dumps({'filePath': filepath})
    destination = '/topic/org.dawnsci.file.topic'
    conn.send(destination, message, ack='auto')
    print('Message sent!')


"""
A python editor
"""

import tkinter as tk
from tkinter import messagebox, filedialog
from mmg_toolbox.tkguis.misc.config import get_config, C
from mmg_toolbox.tkguis.misc.functions import topmenu
from mmg_toolbox.tkguis.misc.styles import create_root, RootWithStyle
from mmg_toolbox.tkguis.misc.logging import create_logger


logger = create_logger(__file__)


def create_python_editor(script_string: str | None = None, parent: tk.Misc | None = None,
                         config: dict | None = None, filename: str | None = None) -> RootWithStyle:
    """
    ScriptGenerator GUI
    Python code editor with special features for I16 scripts

    :param script_string: string to display
    :param parent: tkinter parent object
    :param config: config dict
    :param filename: initial filename
    """
    from ..widgets.python_editor import PythonEditorFrame

    title = filename or 'Python Editor'
    root = create_root(title, parent=parent)
    config = get_config() if config is None else config

    widget = PythonEditorFrame(root, script_string, config, filename=filename)

    menu = {
        'File': {
            'New script': widget.new,
            'Open': widget.open,
            'Save As...': widget.saveas,
            'Save': widget.save,
            'Quit': root.destroy,
        },
    }
    topmenu(root, menu, add_about=True)

    if parent is None:
        root.mainloop()
    return root




"""
mmg_toolbox tests
Test experiment folder functions
"""

import mmg_toolbox.beamline_metadata.config as config
from mmg_toolbox.beamline_metadata.hdfmap_generic import HdfMapXASMetadata as Md
from . import only_dls_file_system
from .example_files import DIR


def test_config():
    cfg = config.beamline_config('i16')
    assert cfg[config.C.beamline] == 'i16'
    assert 'Atten' in cfg[config.C.metadata_string]

    cfg = config.beamline_config('i06-1')
    assert cfg[config.C.beamline] == 'i06-1'
    assert 'field =' in cfg[config.C.metadata_string]
    assert cfg[config.C.default_directory].startswith('/dls/i06-1/data/')

    cfg = config.beamline_config('i32')
    assert cfg[config.C.beamline] == 'i32'

    cfg = config.beamline_config()
    assert len(cfg[config.C.roi]) == 0


@only_dls_file_system
def test_metadata_names():
    import hdfmap
    f = DIR + 'i10/i10-1-26851.nxs'
    m = hdfmap.NexusMap()
    expressions = (name for name in dir(Md) if not name.startswith('_'))
    with hdfmap.load_hdf(f) as h:
        m.populate(h)
        for expression in expressions:
            print(expression, getattr(Md, expression))
            value = m.eval(h, getattr(Md, expression), default=None)
            print(expression, getattr(Md, expression), value)
            assert value is not None




"""
Experiment Folder Monitor
"""

import os
import numpy as np
import hdfmap

from ..utils.misc_functions import numbers2string
from ..utils.env_functions import scan_number_mapping, last_folder_update
from ..nexus.nexus_scan import NexusScan, NexusDataHolder
from ..nexus.nexus_reader import find_scans
from ..xas import load_xas_scans, SpectraContainer

DEFAULT_SCAN_DESCRIPTION = '{(cmd|scan_command)}'


class Experiment:
    """
    Experiment class
    Monitors data folders for scans
    """
    _scan_description = DEFAULT_SCAN_DESCRIPTION

    def __init__(self, *folder_paths: str, instrument: str | None = None):
        self.folder_paths = folder_paths
        self.scan_list = {}
        self._scan_list_update = None
        self.instrument = instrument
        from ..plotting.exp_plot_manager import ExperimentPlotManager
        self.plot = ExperimentPlotManager(self)

    def __repr__(self):
        paths = ', '.join("'{p}'" for p in self.folder_paths)
        return f"Experiment({paths}, instrument={self.instrument})"

    def __str__(self):
        self._update_scan_list()
        scan_numbers = self._scan_numbers()
        lines = ['Instrument: ' + self.instrument]
        lines.extend(self.folder_paths)
        lines.extend([
            f"    Files: {len(scan_numbers)}",
            f"    Scans: {scan_numbers[0]}-{scan_numbers[-1]}",
        ])
        return '\n'.join(lines)

    def _update_scan_list(self):
        mod_times = [last_folder_update(folder) for folder in self.folder_paths]
        folders = [
            folder for folder, time in zip(self.folder_paths, mod_times)
            if self._scan_list_update is None or time > self._scan_list_update
        ]
        self._scan_list_update = max(mod_times)  # datetime.now()?
        self.scan_list.update(scan_number_mapping(*folders))

    def _scan_numbers(self) -> list[int]:
        self._update_scan_list()
        return list(self.scan_list.keys())

    def add_data_paths(self, *folder_paths: str):
        """Add additional paths"""
        new_paths = tuple(
            path for path in folder_paths
            if path not in self.folder_paths and os.path.isdir(path)
        )
        self.folder_paths = self.folder_paths + new_paths

    def all_scans(self) -> dict[int, str]:
        self._update_scan_list()
        return self.scan_list.copy()

    def all_scan_numbers(self) -> list[int]:
        self._update_scan_list()
        return list(self.scan_list.keys())

    def get_scan_filename(self, scan_file: int | str = -1) -> str:
        """Return the full filename of a scan number"""
        if isinstance(scan_file, int):
            if scan_file < 1:
                scan_numbers = self._scan_numbers()
                return self.scan_list[scan_numbers[scan_file]]
            self._update_scan_list()
            return self.scan_list[scan_file]

        if os.path.isfile(scan_file):
            return os.path.abspath(scan_file)
        raise FileNotFoundError(f"scan file {scan_file} not found")

    def scan(self, scan_file: int | str = -1) -> NexusDataHolder:
        """read Nexus file as NexusDataHolder"""
        return NexusDataHolder(self.get_scan_filename(scan_file))

    def scans(self, *scan_files: int | str, hdf_map: hdfmap.NexusMap | None = None) -> list[NexusScan]:
        """Read Nexus files as NexusScan"""
        filenames = [self.get_scan_filename(scan_file) for scan_file in scan_files]
        if not filenames:
            filenames = list(self.all_scans().values())
        if filenames and hdf_map is None:
            hdf_map = hdfmap.create_nexus_map(filenames[0])
        return [NexusScan(file, hdf_map) for file in filenames]

    def find_scans(self, *scan_files: int | str,  hdf_map: hdfmap.NexusMap | None = None, first_only: bool = False,
                   **matches: str | float | tuple[float, float]) -> list[NexusScan]:
        """
        Find scans files with matching metadata

            matches = {
                'name1': 'scan', # matches if 'scan' in file['name1']
                'name2': value, # matches if file['name2'] ~= value
                'name3': (value, tol), # matches if abs(file['name3'] - value) < tol
            }
            match_files = find_scans(*filenames, **matches)

        :param scan_files: set of scan numbers or filenames, if not given will search all scans in folder.
        :param hdf_map: if given, uses this hdfmap rather than generating one.
        :param first_only: if true, returns on the first result
        :param matches: keyword arguments for matching parameters
        :returns: list of scan files that match all requirements
        """
        filenames = [self.get_scan_filename(scan_file) for scan_file in scan_files]
        if not filenames:
            filenames = list(self.all_scans().values())
        if hdf_map is None:
            hdf_map = hdfmap.create_nexus_map(filenames[0])
        matches = find_scans(*filenames, hdf_map=hdf_map, first_only=first_only, **matches)
        return self.scans(*matches, hdf_map=hdf_map)

    def join_scan_data(self, *scan_files: int | str, hdf_map: hdfmap.NexusMap | None = None,
                       data_fields: list[str] | None = None, default: np.ndarray = np.array([0.0])) -> dict[str, list]:
        """
        Join data from scans
        """
        scans = self.scans(*scan_files, hdf_map=hdf_map)
        data_fields = [DEFAULT_SCAN_DESCRIPTION] if data_fields is None else data_fields
        data = {name: [] for name in data_fields}
        for scan in scans:
            with scan.load_hdf() as hdf:
                for name in data_fields:
                    data[name].append(scan.map.eval(hdf, name, default=default))
        return data

    def get_all_data(self, *fields: str, default: np.ndarray = np.array([0.0])) -> dict[str, list]:
        """
        Return dict of data for all files
        """
        return self.join_scan_data(data_fields=list(fields), default=default)

    def generate_mesh(self, *scan_files: int | str, hdf_map: hdfmap.NexusMap | None = None,
                      axes: str | tuple[str, str] = 'axes', signal: str = 'axes',
                      values: str | None = None) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Generate 2D mesh from scan or scans

            x, y, z = generate_mesh(*range(-10, 0), axes='eta', signal='roi2_sum', values='Tsample')
            # or, if scan 12345 is a 2D grid scan
            x, y, z = generate_mesh(12345, axes=('sx', 'sy'), signal='roi2_sum')

        :param scan_files: multiple files or single 2D grid scan
        :param hdf_map: hdfmap.NeXus map, or None to generate
        :param axes: x-axis name, or for grid scans the names of ('xaxis', 'yaxis')
        :param signal: signal name
        :param values: name of the value that changes between multiple files
        :returns: X, Y, IMAGE rank 2 arrays
        """
        scans = self.scans(*scan_files, hdf_map=hdf_map)
        if len(scan_files) == 1:
            # Single 2D Grid scan
            scan = scans[0]
            try:
                x_axis, y_axis = axes
            except ValueError:
                raise Exception("axes should be specified as axes=('axes0', 'axes1')")
            with scan.load_hdf() as hdf:
                x_data = scan.map.eval(hdf, x_axis, default=0.0).squeeze()
                y_data = scan.map.eval(hdf, y_axis, default=0.0).squeeze()
                z_data = scan.map.eval(hdf, signal, default=0.0).squeeze()
            if x_data.size != y_data.size:
                raise Exception(f"arrays '{x_axis}'[{x_data.size}] and '{y_axis}'[{y_data.size}] have different sizes")
            if x_data.ndim == 1 and y_data.ndim == 1:
                y_data, x_data = np.meshgrid(y_data, x_data)
            if z_data.shape != x_data.shape or z_data.shape != y_data.shape:
                raise Exception(
                    f"{repr(scan)} '{x_axis}', '{y_axis}' and '{signal}' shapes are not consistent: " +
                    f"x: {x_data.shape}, y: {y_data.shape}, im: {z_data.shape}"
                )
            return x_data, y_data, z_data
        else:
            x_data, y_data, z_data = [], [], []
            for n, scan in enumerate(scans):
                with scan.load_hdf() as hdf:
                    x = np.reshape(scan.map.eval(hdf, axes, default=0.0), -1)
                    y = np.reshape(scan.map.eval(hdf, signal, default=0.0), -1)
                    val = np.reshape(scan.map.eval(hdf, values, default=0.0), -1) if values is not None else np.array([n])
                if val.size == 1:
                    val = np.tile(val, x.size)
                if y.size != x.size or val.size != x.size:
                    raise Exception(
                        f"{repr(scan)} '{axes}', '{signal}' and '{values}' shapes are not consistent: " +
                        f"x, y, val = ({x.size}, {y.size}, {val.size})"
                    )
                x_data.append(x)
                y_data.append(val)
                z_data.append(y)
            # create a regular sized array
            min_len = min(len(x) for x in x_data)
            # array size [len(scan_files), min_len]
            x_array = np.array([x[:min_len] for x in x_data])
            y_array = np.array([y[:min_len] for y in y_data])
            z_array = np.array([z[:min_len] for z in z_data])
            return x_array, y_array, z_array

    def scan_str(self, scan_file: int | str = 0, metadata_str: str | None = None,
                 hdf_map: hdfmap.NexusMap | None = None) -> str:
        """Read scan file and return metadata string"""
        if metadata_str is None:
            from ..tkguis.misc.beamline_metadata import META_STRING, BEAMLINE_META
            if self.instrument in BEAMLINE_META:
                metadata_str = BEAMLINE_META[self.instrument]
            else:
                metadata_str = META_STRING

        scan_file = self.get_scan_filename(scan_file)

        if hdf_map is None:
            hdf_map = hdfmap.create_nexus_map(scan_file)

        with hdfmap.load_hdf(self.get_scan_filename(scan_file)) as hdf:
            return hdf_map.format_hdf(hdf, metadata_str, raise_errors=True)

    def scans_str(self, *scan_files: int | str, metadata_str: str | None = None,
                  hdf_map: hdfmap.NexusMap | None = None) -> list[str]:
        """Return string description for multiple files"""
        if metadata_str is None:
            metadata_str = " : {str(start_time):30} : " + self._scan_description
        filenames = [self.get_scan_filename(scan_file) for scan_file in scan_files]
        if hdf_map is None:
            hdf_map = hdfmap.create_nexus_map(filenames[0])
        folder_file = ['/'.join(filename.split(os.sep)[-2:]) for filename in filenames]
        return [
            name + self.scan_str(file, metadata_str, hdf_map)
            for file, name in zip(filenames, folder_file)
        ]

    def _generate_scans_title(self, *scans: NexusScan, metadata_str: str | None = None) -> str:
        """Generate title from multiple scan files"""
        if metadata_str is None:
            metadata_str = "\n" + self._scan_description
        first_scan = scans[0]
        folder = first_scan.filename.split(os.sep)[-2]
        meta = first_scan.format(metadata_str)
        scan_numbers = [scan.scan_number() for scan in scans]
        number_range = numbers2string(scan_numbers)
        return f"{folder} {number_range} {meta}"

    def generate_scans_title(self, *scan_files: int | str, metadata_str: str | None = None,
                             hdf_map: hdfmap.NexusMap | None = None) -> str:
        """Generate title from multiple scan files"""
        scans = self.scans(*scan_files, hdf_map=hdf_map)
        return self._generate_scans_title(*scans, metadata_str=metadata_str)

    def load_xas(self, *scan_files: int | str, sample_name: str | None = '') -> list[SpectraContainer]:
        """Read XAS spectra containers"""
        filenames = [self.get_scan_filename(file) for file in scan_files]
        return load_xas_scans(*filenames, sample_name=sample_name)




"""
mmg_toolbox tests
Test utilities
"""
import numpy as np

from mmg_toolbox.diffraction import lattice
from mmg_toolbox.utils import xray_utils, rotations, misc_functions, units


def test_wavelength():
    assert abs(xray_utils.photon_energy(1.0) - 12.398) < 0.01
    assert abs(xray_utils.photon_wavelength(8) - 1.55) < 0.01
    assert abs(lattice.wavevector(1.0) - 6.283) < 0.01


def test_rotation():
    rotation = rotations.rotation_t_matrix(30, (1, 1, 0))
    translation = rotations.translation_t_matrix(5, (0, 0, 1))
    total = translation @ rotation
    vec = (1, 1, 1)
    new_vec1 = rotations.transform_by_t_matrix(vec, rotation)
    new_vec2 = rotations.transform_by_t_matrix(new_vec1, translation)
    new_vec3 = rotations.transform_by_t_matrix(vec, total)
    assert np.allclose(new_vec2, new_vec3)

    phi, eta, chi, mu = 0, 25.5 / 2, 90, 0
    r1 = rotations.rotmatrix_diffractometer(phi, chi, eta, mu)
    r2 = np.dot(
        rotations.rotmatrix_x(mu), np.dot(
            rotations.rotmatrix_z(-eta), np.dot(
                rotations.rotmatrix_y(chi), rotations.rotmatrix_z(-phi)
            )
        )
    )
    assert np.sum(np.abs(r1 - r2)) < 0.01


def test_lattice_orientation():
    np.set_printoptions(precision=3, suppress=True)
    phi, eta, chi, mu = 0, 25.5 / 2, 90, 0
    a, b, c, alpha, beta, gamma = 2.85, 2.85, 10.8, 90, 90, 120.
    b_matrix = lattice.bmatrix(a, b, c, alpha, beta, gamma)
    u_matrix = np.eye(3)
    r_matrix = rotations.rotmatrix_diffractometer(phi, chi, eta, mu)
    lab_transform = np.array([[0, 0, 1], [1, 0, 0], [0, 1, 0]])  # Diffcalc -> DLS
    avec, bvec, cvec = np.linalg.inv(b_matrix)
    # example magnetic moment
    mx, my, mz = 0, 0, 3
    momentmag = np.sqrt(np.sum(np.square([mx, my, mz])))
    momentxyz = np.dot([mx, my, mz], [avec, bvec, cvec])
    moment = momentmag * momentxyz / np.sqrt(np.sum(np.square(momentxyz)))  # broadcast n*1 x n*3 = n*3
    moment[np.isnan(moment)] = 0.
    # Convert to lab coordinates
    moment_lab1 = np.dot(lab_transform, np.dot(r_matrix, np.dot(u_matrix, moment)))
    print(f"\n\nMoment in lab coordinates (in uB): {moment_lab1}")

    # alternative approach
    ub_matrix = np.dot(u_matrix, b_matrix)
    ub_rl_matrix = np.dot(lab_transform, np.dot(r_matrix, ub_matrix))
    momentmag = np.sqrt(np.sum(np.square([mx, my, mz])))
    momentxyz = np.dot([mx, my, mz], np.linalg.inv(ub_rl_matrix))
    moment_lab2 = momentmag * momentxyz / np.sqrt(np.sum(np.square(momentxyz)))  # broadcast n*1 x n*3 = n*3
    moment_lab2[np.isnan(moment_lab2)] = 0.
    print(f"Moment in lab coordinates (in uB): {moment_lab2}")
    momentuvw = np.linalg.norm(moment_lab2) * np.dot(moment_lab2, ub_rl_matrix) / np.linalg.norm(np.dot(moment_lab2, ub_rl_matrix))
    print(f"Moment in crystal (in uB): {momentuvw}")
    assert np.sum(np.abs(moment_lab1 - moment_lab2)) < 0.001
    assert np.sum(np.abs(momentuvw - [mx, my, mz])) < 0.001


def test_scattering_plane():
    energy_kev = 8
    hkl = [1, 0, 6]
    a, b, c, alpha, beta, gamma = 2.85, 2.85, 10.8, 90, 90, 120.

    b_matrix = lattice.bmatrix(a, b, c, alpha, beta, gamma)
    wl = xray_utils.photon_wavelength(energy_kev)
    dspace = 1 / np.linalg.norm(np.dot(hkl, b_matrix))
    tth = lattice.bragg_wl(wl, dspace)
    assert abs(tth - 33.743) < 0.001
    tth2 = lattice.bragg(dspace, energy=energy_kev, en_units='keV')
    assert abs(tth2 - tth) < 0.001

    det_delta, det_gamma = tth, 0
    ki = xray_utils.wavevector_i(wl)
    kf = xray_utils.wavevector_f(wl, det_delta, det_gamma)
    q = kf - ki

    wv_transfer = xray_utils.wavevector_t(wl, det_delta, det_gamma)
    assert np.sum(np.abs(wv_transfer - q)) < 0.001

def test_units():
    distance = units.unit_converter(10, 'nm', 'A')
    assert distance == 100
    energy = units.unit_converter(8, 'keV', 'eV')
    assert energy == 8000


def test_misc_functions():
    assert misc_functions.stfm(35.25, 0.01) == '35.25 (1)'
    assert misc_functions.stfm(110.25, 5) == '110 (5)'
    assert misc_functions.stfm(1.5632e6,1.53e4) == '1.56(2)E+6'

    # misc_functions.numbers2string()





import os
import tkinter as tk
from mmg_toolbox.tkguis.misc.styles import create_root


def create_log_viewer(filename: str, parent: tk.Misc | None = None):
    """Log Viewer Window"""
    from ..widgets.log_viewer import LogViewerWidget
    from mmg_toolbox.utils.file_reader import read_gda_terminal_log

    root = create_root(window_title='Log Viewer', parent=parent)
    log_tabs = read_gda_terminal_log(filename)
    LogViewerWidget(root, log_tabs)
    root.mainloop()
    return root


def create_gda_terminal_log_viewer(data_directory: str, parent: tk.Misc | None = None):
    """Create Log Viewer window using data directory"""
    gda_terminal_log = os.path.join(data_directory, 'gdaterminal.log')
    if os.path.isfile(gda_terminal_log):
        create_log_viewer(gda_terminal_log, parent=parent)
    else:
        raise OSError(f"{gda_terminal_log} does not exist!")



"""
NeXus Tools
"""

from .nexus_scan import NexusScan, NexusDataHolder
from .nexus_reader import read_nexus_file, read_nexus_files, find_matching_scans

__all__ = ['NexusScan', 'NexusDataHolder', 'read_nexus_file', 'read_nexus_files', 'find_matching_scans']

